"""
APRON Taylor1+
==============

:Author: Caterina Urban
"""
from _ctypes import POINTER

from apronpy.abstract1 import PyAbstract1
from apronpy.cdll import libt1pD, libt1pMPQ, libt1pMPFR
from apronpy.manager import Manager

libt1pD.t1p_manager_alloc.restype = POINTER(Manager)
libt1pMPQ.t1p_manager_alloc.restype = POINTER(Manager)
libt1pMPFR.t1p_manager_alloc.restype = POINTER(Manager)


class PyT1pD(PyAbstract1):

    manager: POINTER(Manager) = libt1pD.t1p_manager_alloc()


class PyT1pMPQ(PyAbstract1):

    manager: POINTER(Manager) = libt1pMPQ.t1p_manager_alloc()


class PyT1pMPFR(PyAbstract1):

    manager: POINTER(Manager) = libt1pMPFR.t1p_manager_alloc()



"""
APRON Polyhedra
===============

:Author: Caterina Urban
"""
from _ctypes import POINTER
from abc import ABCMeta

from apronpy.abstract1 import PyAbstract1
from apronpy.cdll import libpolkaMPQ, libpolkaRll
from apronpy.manager import Manager

libpolkaMPQ.pk_manager_alloc.restype = POINTER(Manager)
libpolkaRll.pk_manager_alloc.restype = POINTER(Manager)


class PyPolkaMPQ(PyAbstract1, metaclass=ABCMeta):
    pass


class PyPolkaMPQloose(PyPolkaMPQ):

    manager: POINTER(Manager) = libpolkaMPQ.pk_manager_alloc(False)


class PyPolkaMPQstrict(PyPolkaMPQ):

    manager: POINTER(Manager) = libpolkaMPQ.pk_manager_alloc(True)


class PyPolkaRll(PyAbstract1, metaclass=ABCMeta):
    pass


class PyPolkaRllloose(PyPolkaRll):

    manager: POINTER(Manager) = libpolkaRll.pk_manager_alloc(False)


class PyPolkaRllstrict(PyPolkaRll):

    manager: POINTER(Manager) = libpolkaRll.pk_manager_alloc(True)



"""
APRON Linear Constraints (Level 0)
==================================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER
from enum import IntEnum

from apronpy.cdll import libapron
from apronpy.linexpr0 import Linexpr0
from apronpy.scalar import Scalar, c_uint, c_size_t


class ConsTyp(IntEnum):
    """
    typedef enum ap_constyp_t {
      AP_CONS_EQ,    /* equality constraint */
      AP_CONS_SUPEQ, /* >= constraint */
      AP_CONS_SUP,   /* > constraint */
      AP_CONS_EQMOD, /* congruence equality constraint */
      AP_CONS_DISEQ  /* disequality constraint */
    } ap_constyp_t;
    """
    AP_CONS_EQ = 0
    AP_CONS_SUPEQ = 1
    AP_CONS_SUP = 2
    AP_CONS_EQMOD = 3
    AP_CONS_DISEQ = 4

    def __repr__(self):
        if self.value == 0:
            return '=='
        elif self.value == 1:
            return '>='
        elif self.value == 2:
            return '>'
        elif self.value == 3:
            return '%='
        elif self.value == 4:
            return '!='
        return ValueError('no matching constraint type')


class Lincons0(Structure):
    """
    typedef struct ap_lincons0_t {
      ap_linexpr0_t* linexpr0;  /* expression */
      ap_constyp_t constyp;     /* type of constraint */
      ap_scalar_t* scalar;      /* maybe NULL. For EQMOD constraint, indicates the modulo */
    } ap_lincons0_t;
    """

    _fields_ = [
        ('linexpr0', POINTER(Linexpr0)),
        ('constyp', c_uint),
        ('scalar', POINTER(Scalar))
    ]


class Lincons0Array(Structure):
    """
    typedef struct ap_lincons0_array_t {
      ap_lincons0_t* p;
      size_t size;
    } ap_lincons0_array_t;
    """

    _fields_ = [
        ('p', POINTER(Lincons0)),
        ('size', c_size_t)
    ]


libapron.ap_lincons0_is_unsat.argtypes = [POINTER(Lincons0)]



#
import faulthandler
from _ctypes import POINTER

from apronpy.tcons1 import PyTcons1
from memory_profiler import profile

from apronpy.cdll import libapron
from apronpy.coeff import PyMPQScalarCoeff
from apronpy.environment import PyEnvironment
from apronpy.lincons0 import ConsTyp
from apronpy.lincons1 import PyLincons1, PyLincons1Array
from apronpy.linexpr1 import PyLinexpr1
from apronpy.polka import PyPolkaMPQstrict
from apronpy.texpr0 import Texpr0
from apronpy.texpr1 import PyTexpr1
from apronpy.var import PyVar

faulthandler.enable()

# libapron.ap_texpr1_print.argtypes = [POINTER(Texpr1)]
# libapron.ap_texpr1_print(t.texpr1)
# print()

libapron.ap_texpr0_print.argtypes = [POINTER(Texpr0)]


def main():
    # x = PyVar('o1')
    # e = PyEnvironment([], [x, PyVar('o2')])
    # # # outcome: o1 < o2
    # o1 = PyTexpr1.var(e, PyVar('o1'))
    # o2 = PyTexpr1.var(e, PyVar('o2'))
    # x0 = PyTexpr1.binop(TexprOp.AP_TEXPR_SUB, o2, o1, TexprRtype.AP_RTYPE_REAL,
    #                     TexprRdir.AP_RDIR_RND)
    # c0 = PyTcons1.make(x0, ConsTyp.AP_CONS_SUP)
    # lst = [c0]
    # # del c0
    # a0 = PyTcons1Array(lst)
    # p0 = PyPolkaMPQstrict(e)
    # p = p0.meet(a0)
    # # o1 = 4 * h1 -1 * h2 + 4
    # # o2 = -0.2 * h1 + 0.3 * h2 + 0.4
    # h1 = PyTexpr1.var(e, PyVar('h1'))
    # h2 = PyTexpr1.var(e, PyVar('h2'))
    # w1 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
    # w2 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.1))
    # b1 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
    # c1 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w1, h1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # c2 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w2, h2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # x0 = PyTexpr1.binop(TexprOp.AP_TEXPR_SUB, c1, c2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # x2a = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x0, b1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # w3 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.2))
    # w4 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.3))
    # b2 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
    # c3 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w3, h1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # c4 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w4, h2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # x1 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, c3, c4, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # x2b = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x1, b2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
    # p1 = p.substitute(PyVar('o2'), x2b)
    # r = p1.substitute(PyVar('o1'), x2a)

    size = 1#00000

    for i in range(size):
        e = PyEnvironment([PyVar('x0'), PyVar('y')], [PyVar('z')])
        x0 = PyLinexpr1(e)
        x0.set_coeff(PyVar('x0'), PyMPQScalarCoeff(1))
        x0.set_cst(PyMPQScalarCoeff(3))
        t0 = PyTexpr1(x0)
        c0 = PyTcons1.make(t0, ConsTyp.AP_CONS_SUPEQ)
        print(c0)  # '3 + 1 · x0 >= 0'
        x1 = PyLinexpr1(e)
        x1.set_coeff(PyVar('x0'), PyMPQScalarCoeff(1))
        x1.set_cst(PyMPQScalarCoeff(-1))
        t1 = PyTexpr1(x1)
        c1 = PyTcons1.make(t1, ConsTyp.AP_CONS_SUPEQ)
        print(c1)  # '-1 + 1 · x0 >= 0'
        s = c0.substitute(PyVar('x0'), t1)
        print(s)

        # z = PyMPZ(3)
        # q = PyMPQ(3)
        # del z
        # del q
        #
        # x = PyVar('o1')
        # e = PyEnvironment([], [x, PyVar('o2')])
        # # # outcome: o1 < o2
        # o1 = PyTexpr1.var(e, PyVar('o1'))
        # o2 = PyTexpr1.var(e, PyVar('o2'))
        # x0 = PyTexpr1.binop(TexprOp.AP_TEXPR_SUB, o2, o1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # c0 = PyTcons1.make(x0, ConsTyp.AP_CONS_SUP)
        # cp = deepcopy(c0)
        # lst = [c0, cp]
        # # del c0
        # a0 = PyTcons1Array(lst)



        #
        #
        # del a0
        # # o1 = 4 * h1 -1 * h2 + 4
        # # o2 = -0.2 * h1 + 0.3 * h2 + 0.4
        # h1 = PyTexpr1.var(e, PyVar('h1'))
        # h2 = PyTexpr1.var(e, PyVar('h2'))
        # w1 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # w2 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.1))
        # b1 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # c1 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w1, h1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # c2 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w2, h2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # x0 = PyTexpr1.binop(TexprOp.AP_TEXPR_SUB, c1, c2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # x2a = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x0, b1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # w3 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.2))
        # w4 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.3))
        # b2 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
        # c3 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w3, h1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # c4 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w4, h2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # x1 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, c3, c4, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # x2b = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x1, b2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
        # p1 = p.substitute(PyVar('o2'), x2b)
        # r = p1.substitute(PyVar('o1'), x2a)
        # del r



# e = PyEnvironment([], [
#     PyVar('i1'), PyVar('i2'),
#     PyVar('b1'), PyVar('h1'),
#     PyVar('h2'), PyVar('b2'),
#     PyVar('o1'), PyVar('o2'),
# ])
# # outcome: o1 < o2
# x0 = PyLinexpr1(e)
# x0.set_coeff(PyVar('o1'), PyMPQScalarCoeff(-1))
# x0.set_coeff(PyVar('o2'), PyMPQScalarCoeff(1))
# c0 = PyLincons1(ConsTyp.AP_CONS_SUP, x0)
# print('c0', c0)
# o1 = PyTexpr1.var(e, PyVar('o1'))
# o2 = PyTexpr1.var(e, PyVar('o2'))
# x0bis = PyTexpr1.binop(TexprOp.AP_TEXPR_SUB, o2, o1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# c0bis = PyTcons1.make(x0bis, ConsTyp.AP_CONS_SUP)
# print('c0bis', c0bis)
#
# a0 = PyLincons1Array([c0])
# p0 = PyPolkaMPQstrict(e).meet(a0)
# p0bis = PyPolkaMPQstrict(e).meet(a0)
# print('p0/p0bis', p0, p0bis)
# a0bis = PyTcons1Array([c0bis.tcons1])
# p0bisbis = PyPolkaMPQstrict(e).meet(a0bis)
# print('p0bisbis', p0bisbis)
#
# # o1 = 4 * h1 -1 * h2 + 4
# # o2 = -0.2 * h1 + 0.3 * h2 + 0.4
# h1 = PyTexpr1.var(e, PyVar('h1'))
# h2 = PyTexpr1.var(e, PyVar('h2'))
# w1 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
# w2 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.1))
# b1 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
# c1 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w1, h1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# c2 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w2, h2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# x0 = PyTexpr1.binop(TexprOp.AP_TEXPR_SUB, c1, c2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# x2a = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x0, b1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# w3 = PyTexpr1.cst(e, PyMPQScalarCoeff(-0.2))
# w4 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.3))
# b2 = PyTexpr1.cst(e, PyMPQScalarCoeff(0.4))
# c3 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w3, h1, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# c4 = PyTexpr1.binop(TexprOp.AP_TEXPR_MUL, w4, h2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# x1 = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, c3, c4, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# x2b = PyTexpr1.binop(TexprOp.AP_TEXPR_ADD, x1, b2, TexprRtype.AP_RTYPE_REAL, TexprRdir.AP_RDIR_RND)
# p1 = p0.substitute(PyVar('o2'), x2b)
# p2 = p1.substitute(PyVar('o1'), x2a)
# print('x2b', x2b)
# p2bis = p0bis.substitute([PyVar('o1'), PyVar('o2')], [x2a, x2b])
# p1bis = p0bisbis.substitute(PyVar('o2'), x2b)
# print(p1, p2, p2bis, p1bis)

#(- 3602879701896397/18014398509481984) · h1 + 5404319552844595/18014398509481984 · h2 + 3602879701896397/9007199254740992

@profile
def test():
    e = PyEnvironment([], [
        PyVar('i1'), PyVar('i2'),
        PyVar('b1'), PyVar('h1'),
        PyVar('h2'), PyVar('b2'),
        PyVar('o1'), PyVar('o2')
    ])
    result = PyPolkaMPQstrict.bottom(e)

    for i in range(4):
        for j in range(4):

            # outcome: o1 < o2
            x0 = PyLinexpr1(e)
            x0.set_coeff(PyVar('o1'), PyMPQScalarCoeff(-1))
            x0.set_coeff(PyVar('o2'), PyMPQScalarCoeff(1))
            c0 = PyLincons1(ConsTyp.AP_CONS_SUP, x0)
            a0 = PyLincons1Array([c0])
            p0 = PyPolkaMPQstrict(e).meet(a0)
            p0bis = PyPolkaMPQstrict(e)

            print('Output')
            print('p0: {}'.format(p0))
            print('p0bis: {}'.format(p0bis))
            print('p0 <= p0bis (expected: True): {}'.format(p0 <= p0bis))
            print('p0bis <= p0 (expected: False): {}'.format(p0bis <= p0))
            print()

            if i == 0:
                # relu: assumption 0 <= o1, 0 <= o2
                x1a = PyLinexpr1(e)
                x1a.set_coeff(PyVar('o1'), PyMPQScalarCoeff(1))
                c1a = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x1a)  # 0 <= o1
                x1b = PyLinexpr1(e)
                x1b.set_coeff(PyVar('o2'), PyMPQScalarCoeff(1))
                c1b = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x1b)  # 0 <= o2
                a1 = PyLincons1Array([c1a, c1b])
                p1 = p0.meet(a1)
                p1bis = p0bis.meet(a1)
            elif i == 1:
                # relu: o2 = 0, assumption 0 <= o1, o2 < 0
                x1a = PyLinexpr1(e)
                x1a.set_cst(PyMPQScalarCoeff(0.0))
                p1a = p0.substitute(PyVar('o2'), x1a)
                p1abis = p0bis.substitute(PyVar('o2'), x1a)
                x1b = PyLinexpr1(e)
                x1b.set_coeff(PyVar('o1'), PyMPQScalarCoeff(1))
                c1b = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x1b)  # 0 <= o1
                x1c = PyLinexpr1(e)
                x1c.set_coeff(PyVar('o2'), PyMPQScalarCoeff(-1))
                c1c = PyLincons1(ConsTyp.AP_CONS_SUP, x1c)  # o2 < 0
                a1 = PyLincons1Array([c1b, c1c])
                p1 = p1a.meet(a1)
                p1bis = p1abis.meet(a1)
            elif i == 2:
                # relu: o1 = 0, assumption o1 < 0, 0 <= o2
                x1a = PyLinexpr1(e)
                x1a.set_cst(PyMPQScalarCoeff(0.0))
                p1a = p0.substitute(PyVar('o1'), x1a)
                p1abis = p0bis.substitute(PyVar('o1'), x1a)
                x1b = PyLinexpr1(e)
                x1b.set_coeff(PyVar('o1'), PyMPQScalarCoeff(-1))
                c1b = PyLincons1(ConsTyp.AP_CONS_SUP, x1b)  # o1 < 0
                x1c = PyLinexpr1(e)
                x1c.set_coeff(PyVar('o2'), PyMPQScalarCoeff(1))
                c1c = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x1c)  # 0 <= o2
                a1 = PyLincons1Array([c1b, c1c])
                p1 = p1a.meet(a1)
                p1bis = p1abis.meet(a1)
            else:
                # relu: o1 = 0, o2 = 0, assumption o1 < 0, o2 < 0
                x1a = PyLinexpr1(e)
                x1a.set_cst(PyMPQScalarCoeff(0.0))
                p1a = p0.substitute(PyVar('o1'), x1a)
                p1abis = p0bis.substitute(PyVar('o1'), x1a)
                x1b = PyLinexpr1(e)
                x1b.set_cst(PyMPQScalarCoeff(0.0))
                p1b = p1a.substitute(PyVar('o2'), x1b)
                p1bbis = p1abis.substitute(PyVar('o1'), x1b)
                x1c = PyLinexpr1(e)
                x1c.set_coeff(PyVar('o1'), PyMPQScalarCoeff(-1))
                c1c = PyLincons1(ConsTyp.AP_CONS_SUP, x1c)  # o1 < 0
                x1d = PyLinexpr1(e)
                x1d.set_coeff(PyVar('o2'), PyMPQScalarCoeff(-1))
                c1d = PyLincons1(ConsTyp.AP_CONS_SUP, x1d)  # o2 < 0
                a1 = PyLincons1Array([c1c, c1d])
                p1 = p1b.meet(a1)
                p1bis = p1bbis.meet(a1)

            print('ReLU(o1, o2) {}:{}'.format(i, j))
            print('p1: {}'.format(p1))
            print('p1bis: {}'.format(p1bis))
            print('p1 <= p1bis (expected: True): {}'.format(p1 <= p1bis))
            print('p1bis <= p1 (expected: False): {}'.format(p1bis <= p1))
            print()

            # o1 = -0.7566400101700294 * h1 -0.7521230014426193 * h2 + 0.6
            # o2 = 66145339249.11405 * h1 + 68702343992.50064 * h2 + 0.6
            x2a = PyLinexpr1(e)
            x2a.set_coeff(PyVar('h1'), PyMPQScalarCoeff(4))
            x2a.set_coeff(PyVar('h2'), PyMPQScalarCoeff(-1))
            x2a.set_cst(PyMPQScalarCoeff(4))
            x2b = PyLinexpr1(e)
            x2b.set_coeff(PyVar('h1'), PyMPQScalarCoeff(-2))
            x2b.set_coeff(PyVar('h2'), PyMPQScalarCoeff(3))
            x2b.set_cst(PyMPQScalarCoeff(4))
            p2a = p1.substitute(PyVar('o1'), x2a)
            p2b = p2a.substitute(PyVar('o2'), x2b)
            p2abis = p1bis.substitute(PyVar('o1'), x2a)
            p2bbis = p2abis.substitute(PyVar('o2'), x2b)

            print('Layer(o1,o2) {}:{}'.format(i, j))
            print('p2b: {}'.format(p2b))
            print('p2bis: {}'.format(p2bbis))
            print('p2b <= p2bbis (expected: True): {}'.format(p2b <= p2bbis))
            print('p2bis <= p2b (expected: False): {}'.format(p2bbis <= p2b))
            print()

            if j == 0:
                # relu: assumption 0 <= h1, 0 <= h2
                x3a = PyLinexpr1(e)
                x3a.set_coeff(PyVar('h1'), PyMPQScalarCoeff(1))
                c3a = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x3a)  # 0 <= h1
                x3b = PyLinexpr1(e)
                x3b.set_coeff(PyVar('h2'), PyMPQScalarCoeff(1))
                c3b = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x3b)  # 0 <= h2
                a3 = PyLincons1Array([c3a, c3b])
                p3 = p2b.meet(a3)
                p3bis = p2bbis.meet(a3)
            elif j == 1:
                # relu: h2 = 0, assumption 0 <= h1, h2 < 0
                x3a = PyLinexpr1(e)
                x3a.set_cst(PyMPQScalarCoeff(0.0))
                p3a = p2b.substitute(PyVar('h2'), x3a)
                p3abis = p2bbis.substitute(PyVar('h2'), x3a)
                x3b = PyLinexpr1(e)
                x3b.set_coeff(PyVar('h1'), PyMPQScalarCoeff(1))
                c3b = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x3b)  # 0 <= h1
                x3c = PyLinexpr1(e)
                x3c.set_coeff(PyVar('h2'), PyMPQScalarCoeff(-1))
                c3c = PyLincons1(ConsTyp.AP_CONS_SUP, x3c)  # h2 < 0
                a3 = PyLincons1Array([c3b, c3c])
                p3 = p3a.meet(a3)
                p3bis = p3abis.meet(a3)
            elif j == 2:
                # relu: h1 = 0, assumption h1 < 0, 0 <= h2
                x3a = PyLinexpr1(e)
                x3a.set_cst(PyMPQScalarCoeff(0.0))
                p3a = p2b.substitute(PyVar('h1'), x3a)
                p3abis = p2bbis.substitute(PyVar('h1'), x3a)
                x3b = PyLinexpr1(e)
                x3b.set_coeff(PyVar('h1'), PyMPQScalarCoeff(-1))
                c3b = PyLincons1(ConsTyp.AP_CONS_SUP, x3b)  # h1 < 0
                x3c = PyLinexpr1(e)
                x3c.set_coeff(PyVar('h2'), PyMPQScalarCoeff(1))
                c3c = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x3c)  # 0 <= h2
                a3 = PyLincons1Array([c3b, c3c])
                p3 = p3a.meet(a3)
                p3bis = p3abis.meet(a3)
            else:
                # relu: h1 = 0, h2 = 0, assumption h1 < 0, h2 < 0
                x3a = PyLinexpr1(e)
                x3a.set_cst(PyMPQScalarCoeff(0.0))
                p3a = p2b.substitute(PyVar('h1'), x3a)
                p3abis = p2bbis.substitute(PyVar('h1'), x3a)
                x3b = PyLinexpr1(e)
                x3b.set_cst(PyMPQScalarCoeff(0.0))
                p3b = p3a.substitute(PyVar('h2'), x3b)
                p3bbis = p3abis.substitute(PyVar('h1'), x3b)
                x3c = PyLinexpr1(e)
                x3c.set_coeff(PyVar('h1'), PyMPQScalarCoeff(-1))
                c3c = PyLincons1(ConsTyp.AP_CONS_SUP, x3c)  # h1 < 0
                x3d = PyLinexpr1(e)
                x3d.set_coeff(PyVar('h2'), PyMPQScalarCoeff(-1))
                c3d = PyLincons1(ConsTyp.AP_CONS_SUP, x3d)  # h2 < 0
                a3 = PyLincons1Array([c3c, c3d])
                p3 = p3b.meet(a3)
                p3bis = p3bbis.meet(a3)

            print('ReLU(h1,h2) {}:{}'.format(i, j))
            print('p3: {}'.format(p3))
            print('p3bis: {}'.format(p3bis))
            print('p3 <= p3bis (expected: True): {}'.format(p3 <= p3bis))
            print('p3is <= p3 (expected: False): {}'.format(p3bis <= p3))
            print()

            # h1 = -3563343016573.1133 * i1 + -7126686033146.326 * i2 + 0.35
            # h2 = -4042063042100.3296 * i1 + -8084126084200.859 * i2 + 0.35
            x4a = PyLinexpr1(e)
            x4a.set_coeff(PyVar('i1'), PyMPQScalarCoeff(1))
            x4a.set_coeff(PyVar('i2'), PyMPQScalarCoeff(-1))
            x4a.set_cst(PyMPQScalarCoeff(1))
            x4b = PyLinexpr1(e)
            x4b.set_coeff(PyVar('i1'), PyMPQScalarCoeff(-1))
            x4b.set_coeff(PyVar('i2'), PyMPQScalarCoeff(2))
            x4b.set_cst(PyMPQScalarCoeff(1))
            p4a = p3.substitute(PyVar('h1'), x4a)
            p4b = p4a.substitute(PyVar('h2'), x4b)
            p4abis = p3bis.substitute(PyVar('h1'), x4a)
            p4bbis = p4abis.substitute(PyVar('h2'), x4b)

            print('Layer(h1,h2) {}:{}'.format(i, j))
            print('p4b: {}'.format(p4b))
            print('p4bis: {}'.format(p4bbis))
            print('p4b <= p4bbis (expected: True): {}'.format(p4b <= p4bbis))
            print('p4bbis <= p4b (expected: False): {}'.format(p4bbis <= p4b))
            print()

            # 0 <= i1 <= 1, 0 <= i2 <= 1
            x5a = PyLinexpr1(e)
            # x5a.set_coeff(PyVar('i1'), PyMPQScalarCoeff(1))
            c5a = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x5a)  # i1 >= 0
            x5b = PyLinexpr1(e)
            # x5b.set_coeff(PyVar('i1'), PyMPQScalarCoeff(-1))
            # x5b.set_cst(PyMPQScalarCoeff(1))
            c5b = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x5b)  # -i1 + 1 >= 0
            x5c = PyLinexpr1(e)
            x5c.set_coeff(PyVar('i2'), PyMPQScalarCoeff(1))
            c5c = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x5c)  # i2 >= 0
            x5d = PyLinexpr1(e)
            x5d.set_coeff(PyVar('i2'), PyMPQScalarCoeff(-1))
            x5d.set_cst(PyMPQScalarCoeff(1))
            c5d = PyLincons1(ConsTyp.AP_CONS_SUPEQ, x5d)  # -i2 + 1 >= 0
            a5 = PyLincons1Array([c5a, c5b, c5c, c5d])
            p5 = p4b.meet(a5)
            p5bis = p4bbis.meet(a5)

            print('Input {}:{}'.format(i, j))
            print('p5: {}'.format(p5))
            print('p5is: {}'.format(p5bis))
            print('p5 <= p5bis (expected: True): {}'.format(p5 <= p5bis))
            print('p5bis <= p5b (expected: False): {}'.format(p5bis <= p5))
            print()
            print()

            result = result.join(p5)

    print('Result: {}'.format(result))
    print()


if __name__ == "__main__":
    main()
    # test()
    # e = PyEnvironment([PyVar('x0'), PyVar('y')], [PyVar('z')])
    # x0 = PyLinexpr1(e)
    # x0.set_coeff(PyVar('x0'), PyMPQScalarCoeff(1))
    # x0.set_cst(PyMPQScalarCoeff(3))
    # t0 = PyTexpr1(x0)
    # c0 = PyTcons1.make(t0, ConsTyp.AP_CONS_SUPEQ)
    # print(c0)   # '3 + 1 · x0 >= 0'
    # x1 = PyLinexpr1(e)
    # x1.set_coeff(PyVar('x0'), PyMPQScalarCoeff(1))
    # x1.set_cst(PyMPQScalarCoeff(-1))
    # t1 = PyTexpr1(x1)
    # c1 = PyTcons1.make(t1, ConsTyp.AP_CONS_SUPEQ)
    # print(c1)   # '-1 + 1 · x0 >= 0'
    # s = c0.substitute(PyVar('x0'), t1)
    # print(s)
    # self.assertEqual(str(c0.substitute(PyVar('x0'), t1)), '3 + 1 · (-1 + 1 · x0) >= 0')

print("Done")



"""
GMP Multi-Precision Integers
============================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from ctypes import c_int, c_ulonglong, c_double, c_char_p, c_size_t, create_string_buffer, c_char
from typing import Union

from apronpy.cdll import libgmp

# initialization and assignment functions
MPZ_clear = libgmp.__gmpz_clear
MPZ_init_set = libgmp.__gmpz_init_set
MPZ_init_set_d = libgmp.__gmpz_init_set_d
# conversion functions
MPZ_get_str = libgmp.__gmpz_get_str
# comparison functions
MPZ_cmp = libgmp.__gmpz_cmp  # -1: op1 < op2, 0: op1 == op2, 1: op1 > op2
# arithmetic functions
MPZ_add = libgmp.__gmpz_add
MPZ_sub = libgmp.__gmpz_sub
MPZ_mul = libgmp.__gmpz_mul
MPZ_neg = libgmp.__gmpz_neg
MPZ_abs = libgmp.__gmpz_abs
MPZ_sizeinbase = libgmp.__gmpz_sizeinbase


class MPZ(Structure):
    """
    typedef struct
    {
      int _mp_alloc;
      int _mp_size;
      mp_limb_t *_mp_d;
    } __mpz_struct;
    """
    _fields_ = [
        ('_mp_alloc', c_int),
        ('_mp_size', c_int),
        ('_mp_d', POINTER(c_ulonglong))
    ]

    def __repr__(self):
        buffer = create_string_buffer(MPZ_sizeinbase(self, 10) + 2)
        return MPZ_get_str(buffer, 10, self).decode('utf-8')


class PyMPZ:

    def __init__(self, value: Union[MPZ, int, float] = 0):
        self.mpz = MPZ()
        if isinstance(value, MPZ):
            MPZ_init_set(self, value)
        else:
            assert isinstance(value, (int, float))
            MPZ_init_set_d(self, c_double(value))

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = PyMPZ(self.mpz)
        memodict[id(self)] = result
        return result

    def __del__(self):
        MPZ_clear(self)
        del self.mpz

    @property
    def _as_parameter_(self):
        return byref(self.mpz)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyMPZ)
        return argument

    def __repr__(self):
        buffer = create_string_buffer(MPZ_sizeinbase(self, 10) + 2)
        return MPZ_get_str(buffer, 10, self.mpz).decode('utf-8')

    def __lt__(self, other: 'PyMPZ'):
        assert isinstance(other, PyMPZ)
        return MPZ_cmp(self, other) < 0

    def __le__(self, other: 'PyMPZ'):
        assert isinstance(other, PyMPZ)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: 'PyMPZ'):
        assert isinstance(other, PyMPZ)
        return MPZ_cmp(self, other) == 0

    def __ne__(self, other: 'PyMPZ'):
        assert isinstance(other, PyMPZ)
        return not self.__eq__(other)

    def __ge__(self, other: 'PyMPZ'):
        assert isinstance(other, PyMPZ)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other: 'PyMPZ'):
        assert isinstance(other, PyMPZ)
        return MPZ_cmp(self, other) > 0

    def __add__(self, other: 'PyMPZ') -> 'PyMPZ':
        assert isinstance(other, PyMPZ)
        mpz = type(self)(0)
        MPZ_add(mpz, self, other)
        return mpz

    def __sub__(self, other: 'PyMPZ') -> 'PyMPZ':
        assert isinstance(other, PyMPZ)
        mpz = type(self)(0)
        MPZ_sub(mpz, self, other)
        return mpz

    def __mul__(self, other: 'PyMPZ') -> 'PyMPZ':
        assert isinstance(other, PyMPZ)
        mpz = type(self)(0)
        MPZ_mul(mpz, self, other)
        return mpz

    def __neg__(self) -> 'PyMPZ':
        mpz = type(self)(0)
        MPZ_neg(mpz, self)
        return mpz

    def __abs__(self) -> 'PyMPZ':
        mpz = type(self)(0)
        MPZ_abs(mpz, self)
        return mpz


# initialization and assignment functions
MPZ_clear.argtypes = [PyMPZ]
MPZ_init_set.argtypes = [PyMPZ, POINTER(MPZ)]
MPZ_init_set_d.argtypes = [PyMPZ, c_double]
# conversion functions
MPZ_get_str.argtypes = [POINTER(c_char), c_int, POINTER(MPZ)]
MPZ_get_str.restype = c_char_p
# comparison functions
MPZ_cmp.argtypes = [PyMPZ, PyMPZ]
# arithmetic functions
MPZ_add.argtypes = [PyMPZ, PyMPZ, PyMPZ]
MPZ_sub.argtypes = [PyMPZ, PyMPZ, PyMPZ]
MPZ_mul.argtypes = [PyMPZ, PyMPZ, PyMPZ]
MPZ_neg.argtypes = [PyMPZ, PyMPZ]
MPZ_abs.argtypes = [PyMPZ, PyMPZ]
MPZ_sizeinbase.argtypes = [POINTER(MPZ), c_int]
MPZ_sizeinbase.restype = c_size_t



"""
APRON Linear Constraints (Level 1)
==================================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from copy import deepcopy
from ctypes import c_char_p, c_size_t
from typing import List, Union

from apronpy.cdll import libapron
from apronpy.coeff import PyDoubleScalarCoeff, CoeffDiscr, PyMPQIntervalCoeff, \
    PyMPFRIntervalCoeff, \
    PyDoubleIntervalCoeff, PyMPQScalarCoeff, PyMPFRScalarCoeff, PyCoeff, Coeff
from apronpy.environment import Environment, PyEnvironment
from apronpy.lincons0 import Lincons0, ConsTyp, Lincons0Array
from apronpy.linexpr0 import LinexprDiscr
from apronpy.linexpr1 import PyLinexpr1
from apronpy.scalar import PyScalar, c_uint, ScalarDiscr
from apronpy.var import PyVar


class Lincons1(Structure):
    """
    typedef struct ap_lincons1_t {
      ap_lincons0_t lincons0;
      ap_environment_t* env;
    } ap_lincons1_t;
    """

    _fields_ = [
        ('lincons0', Lincons0),
        ('env', POINTER(Environment))
    ]

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = Lincons1()
        result.lincons0 = Lincons0()
        result.lincons0.linexpr0 = libapron.ap_linexpr0_copy(self.lincons0.linexpr0)
        result.lincons0.constyp = self.lincons0.constyp
        if self.lincons0.scalar:
            self.lincons0.scalar = libapron.ap_scalar_alloc_set(self.lincons0.scalar)
        else:
            result.lincons0.scalar = None
        self.env.contents.count += 1
        result.env = self.env
        memodict[id(self)] = result
        return result

    def __repr__(self):
        linexpr0 = self.lincons0.linexpr0.contents
        constyp = ConsTyp(self.lincons0.constyp)
        scalar = self.lincons0.scalar
        env = self.env.contents
        result = ''
        if linexpr0.discr == LinexprDiscr.AP_LINEXPR_DENSE:
            result += ' + '.join(
                '{}·{}'.format(linexpr0.p.coeff[i], env.var_of_dim[i].decode('utf-8'))
                for i in range(linexpr0.size)
            )
            result += ' + {}'.format(linexpr0.cst) if result else '{}'.format(linexpr0.cst)
        else:   # self.discr == LinexprDiscr.AP_LINEXPR_SPARSE:
            terms = list()
            for i in range(linexpr0.size):
                coeff = linexpr0.p.linterm[i].coeff
                dim = linexpr0.p.linterm[i].dim.value
                if dim <= linexpr0.size:
                    terms.append('{}·{}'.format(coeff, env.var_of_dim[dim].decode('utf-8')))
            result += ' + '.join(terms)
            result += ' + {}'.format(linexpr0.cst) if result else '{}'.format(linexpr0.cst)
        if scalar:
            return '{} {} {}'.format(result.replace('+ -', '- '), repr(constyp), scalar.contents)
        else:
            return '{} {} 0'.format(result.replace('+ -', '- '), repr(constyp))


class Lincons1Array(Structure):
    """
    typedef struct ap_lincons1_array_t {
      ap_lincons0_array_t lincons0_array;
      ap_environment_t* env;
    } ap_lincons1_array_t;
    """

    _fields_ = [
        ('lincons0_array', Lincons0Array),
        ('env', POINTER(Environment))
    ]

    def __repr__(self):
        env = self.env.contents
        array = list()
        for i in range(self.lincons0_array.size):

            linexpr0 = self.lincons0_array.p[i].linexpr0.contents
            constyp = ConsTyp(self.lincons0_array.p[i].constyp)
            scalar = self.lincons0_array.p[i].scalar

            result = ''
            if linexpr0.discr == LinexprDiscr.AP_LINEXPR_DENSE:
                result += ' + '.join(
                    '{}·{}'.format(linexpr0.p.coeff[j], env.var_of_dim[j].decode('utf-8'))
                    for j in range(linexpr0.size) if str(linexpr0.p.coeff[j]) != '0'
                )
                result += ' + {}'.format(linexpr0.cst) if result else '{}'.format(linexpr0.cst)
            else:  # self.discr == LinexprDiscr.AP_LINEXPR_SPARSE:
                terms = list()
                for j in range(linexpr0.size):
                    coeff = linexpr0.p.linterm[j].coeff
                    dim = linexpr0.p.linterm[j].dim.value
                    if dim <= linexpr0.size:
                        terms.append('{}·{}'.format(coeff, env.var_of_dim[dim].decode('utf-8')))
                result += ' + '.join(terms)
                result += ' + {}'.format(linexpr0.cst) if result else '{}'.format(linexpr0.cst)
            if scalar:
                result = result.replace('+ -', '- ')
                array.append('{} {} {}'.format(result, repr(constyp), scalar.contents))
            else:
                array.append('{} {} 0'.format(result.replace('+ -', '- '), repr(constyp)))
        return ' ∧ '.join(array)


class PyLincons1:

    def __init__(self, typ: ConsTyp, linexpr: PyLinexpr1, scalar: PyScalar = None):
        self.lincons1 = Lincons1()
        linexpr_copy = deepcopy(linexpr.linexpr1)
        self.lincons1.lincons0.linexpr0 = linexpr_copy.linexpr0
        self.lincons1.lincons0.constyp = c_uint(typ)
        if scalar:
            self.lincons1.lincons0.scalar = deepcopy(scalar)
        self.lincons1.env = linexpr_copy.env

    @classmethod
    def unsat(cls, environment: PyEnvironment):
        x = PyLinexpr1(environment)
        x.set_cst(PyDoubleScalarCoeff(-1.0))
        lincons = cls(ConsTyp.AP_CONS_SUPEQ, x)
        return lincons

    def __del__(self):
        libapron.ap_lincons1_clear(self)
        del self.lincons1

    @property
    def _as_parameter_(self):
        return byref(self.lincons1)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyLincons1)
        return argument

    def __repr__(self):
        return '{}'.format(self.lincons1)

    def is_unsat(self):
        return bool(libapron.ap_lincons0_is_unsat(self.lincons1.lincons0))

    def get_typ(self):
        return ConsTyp(self.lincons1.lincons0.constyp)

    def set_typ(self, typ: ConsTyp):
        self.lincons1.lincons0.constyp = c_uint(typ)

    def get_cst(self):
        cst = self.lincons1.lincons0.linexpr0.contents.cst
        if cst.discr == CoeffDiscr.AP_COEFF_INTERVAL:
            if cst.val.interval.contents.inf.contents.discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQIntervalCoeff()
            elif cst.val.interval.contents.inf.contents.discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRIntervalCoeff(0, 0)
            else:  # cst.val.interval.contents.inf.contents.discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleIntervalCoeff()
        else:  # CoeffDiscr.AP_COEFF_SCALAR
            if cst.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQScalarCoeff()
            elif cst.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRScalarCoeff(0)
            else:  # cst.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleScalarCoeff()
        libapron.ap_coeff_set(result.coeff, self.lincons1.lincons0.linexpr0.contents.cst)
        return result

    def set_cst(self, cst: PyCoeff):
        libapron.ap_coeff_set(byref(self.lincons1.lincons0.linexpr0.contents.cst), cst.coeff)

    def get_coeff(self, var: PyVar):
        coeff = libapron.ap_lincons1_coeffref(self, var._as_parameter_)
        if coeff.contents.discr == CoeffDiscr.AP_COEFF_INTERVAL:
            discr = coeff.contents.val.interval.contents.inf.contents.discr
            if discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQIntervalCoeff()
            elif discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRIntervalCoeff(0, 0)
            else:  # discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleIntervalCoeff()
        else:  # CoeffDiscr.AP_COEFF_SCALAR
            if coeff.contents.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQScalarCoeff()
            elif coeff.contents.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRScalarCoeff(0)
            else:  # coeff.contents.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleScalarCoeff()
        libapron.ap_lincons1_get_coeff(result.coeff, self, var._as_parameter_)
        return result

    def set_coeff(self, var: PyVar, coeff: PyCoeff):
        libapron.ap_coeff_set(libapron.ap_lincons1_coeffref(self, var._as_parameter_), coeff.coeff)


libapron.ap_lincons1_clear.argtypes = [PyLincons1]
libapron.ap_lincons1_coeffref.argtypes = [PyLincons1, c_char_p]
libapron.ap_lincons1_coeffref.restype = POINTER(Coeff)
libapron.ap_lincons1_get_coeff.argtypes = [POINTER(Coeff), PyLincons1, c_char_p]


class PyLincons1Array:

    def __init__(self, lincons1s: Union[Lincons1Array, List[PyLincons1]] = None,
                 environment: PyEnvironment = None):
        if isinstance(lincons1s, Lincons1Array):
            self.lincons1array = lincons1s
        elif lincons1s:
            size = len(lincons1s)
            self.lincons1array = libapron.ap_lincons1_array_make(lincons1s[0].lincons1.env, size)
            for i in range(size):
                libapron.ap_lincons1_array_set(self, i, deepcopy(lincons1s[i].lincons1))
        else:
            self.lincons1array = libapron.ap_lincons1_array_make(environment, 0)

    def __del__(self):
        libapron.ap_lincons1_array_clear(self)
        del self.lincons1array

    @property
    def _as_parameter_(self):
        return byref(self.lincons1array)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyLincons1Array)
        return argument

    def __repr__(self):
        return '{}'.format(self.lincons1array)


libapron.ap_lincons1_array_make.argtypes = [POINTER(Environment), c_size_t]
libapron.ap_lincons1_array_make.restype = Lincons1Array
libapron.ap_lincons1_array_clear.argtypes = [PyLincons1Array]
libapron.ap_lincons1_array_set.argtypes = [PyLincons1Array, c_size_t, POINTER(Lincons1)]



"""
APRON Tree Expressions (Level 1)
================================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER
from ctypes import c_char_p, c_int
from typing import Union

from apronpy.cdll import libapron
from apronpy.coeff import PyCoeff
from apronpy.environment import Environment, PyEnvironment
from apronpy.linexpr1 import PyLinexpr1
from apronpy.texpr0 import Texpr0, TexprDiscr, TexprOp, TexprRtype, TexprRdir
from apronpy.var import PyVar


class Texpr1(Structure):
    """
    typedef struct ap_texpr1_t {
      ap_texpr0_t* texpr0;
      ap_environment_t* env;
    } ap_texpr1_t;
    """

    _fields_ = [
        ('texpr0', POINTER(Texpr0)),
        ('env', POINTER(Environment))
    ]

    def __repr__(self):
        def precendence(texpr0):
            op_precedence = {
                TexprOp.AP_TEXPR_ADD: 1,
                TexprOp.AP_TEXPR_SUB: 1,
                TexprOp.AP_TEXPR_MUL: 2,
                TexprOp.AP_TEXPR_DIV: 2,
                TexprOp.AP_TEXPR_MOD: 2,
                TexprOp.AP_TEXPR_POW: 3,
                TexprOp.AP_TEXPR_NEG: 4,
                TexprOp.AP_TEXPR_CAST: 5,
                TexprOp.AP_TEXPR_SQRT: 5
            }
            if texpr0.discr == TexprDiscr.AP_TEXPR_CST or texpr0.discr == TexprDiscr.AP_TEXPR_DIM:
                return op_precedence[TexprOp.AP_TEXPR_NEG]
            return op_precedence[texpr0.val.node.contents.op]

        def do(texpr0, env):
            if texpr0.discr == TexprDiscr.AP_TEXPR_CST:
                return '{}'.format(texpr0.val.cst)
            elif texpr0.discr == TexprDiscr.AP_TEXPR_DIM:
                return '{}'.format(env.var_of_dim[texpr0.val.dim.value].decode('utf-8'))
            else:  # texpr0.discr == TexprDiscr.AP_TEXPR_NODE
                prec = precendence(texpr0)
                prec_a = precendence(texpr0.val.node.contents.exprA.contents)
                if prec_a < prec:
                    expr_a = '({})'.format(do(texpr0.val.node.contents.exprA.contents, env))
                else:
                    expr_a = do(texpr0.val.node.contents.exprA.contents, env)
                op = texpr0.val.node.contents.op
                if texpr0.val.node.contents.exprB:  # binary operation
                    prec_b = precendence(texpr0.val.node.contents.exprB.contents)
                    if prec_b <= prec:
                        expr_b = '({})'.format(do(texpr0.val.node.contents.exprB.contents, env))
                    else:
                        expr_b = do(texpr0.val.node.contents.exprB.contents, env)
                    return '{} {} {}'.format(expr_a, repr(TexprOp(op)), expr_b)
                else:
                    return '{} {}'.format(repr(TexprOp(op)), expr_a)

        return do(self.texpr0.contents, self.env.contents).replace('+ -', '- ')


class PyTexpr1:

    def __init__(self, expr: Union[POINTER(Texpr1), PyLinexpr1]):
        if isinstance(expr, PyLinexpr1):
            self.texpr1: POINTER(Texpr1) = libapron.ap_texpr1_from_linexpr1(expr)
        else:
            assert isinstance(expr, POINTER(Texpr1))
            self.texpr1: POINTER(Texpr1) = expr

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = type(self)(libapron.ap_texpr1_copy(self))
        memodict[id(self)] = result
        return result

    def __del__(self):
        libapron.ap_texpr1_free(self)
        del self.texpr1

    @classmethod
    def cst(cls, environment: PyEnvironment, cst: PyCoeff):
        return cls(libapron.ap_texpr1_cst(environment, cst))

    @classmethod
    def var(cls, environment: PyEnvironment, var: PyVar):
        return cls(libapron.ap_texpr1_var(environment, var))

    @classmethod
    def unop(cls, op: TexprOp, expr, typ: TexprRtype, rdir: TexprRdir):
        return cls(libapron.ap_texpr1_unop(op, libapron.ap_texpr1_copy(expr), typ, rdir))

    @classmethod
    def binop(cls, op: TexprOp, expr_a, expr_b, typ: TexprRtype, rdir: TexprRdir):
        left = libapron.ap_texpr1_copy(expr_a)
        right = libapron.ap_texpr1_copy(expr_b)
        return cls(libapron.ap_texpr1_binop(op, left, right, typ, rdir))

    @property
    def _as_parameter_(self):
        return self.texpr1

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyTexpr1)
        return argument

    def __repr__(self):
        return '{}'.format(self.texpr1.contents)

    def substitute(self, var: PyVar, dst: 'PyTexpr1'):
        return type(self)(libapron.ap_texpr1_substitute(self, var._as_parameter_, dst))


libapron.ap_texpr1_cst.argtypes = [PyEnvironment, PyCoeff]
libapron.ap_texpr1_cst.restype = POINTER(Texpr1)
libapron.ap_texpr1_var.argtypes = [PyEnvironment, c_char_p]
libapron.ap_texpr1_var.restype = POINTER(Texpr1)
libapron.ap_texpr1_unop.argtypes = [c_int, POINTER(Texpr1), c_int, c_int]
libapron.ap_texpr1_unop.restype = POINTER(Texpr1)
libapron.ap_texpr1_binop.argtypes = [c_int, POINTER(Texpr1), POINTER(Texpr1), c_int, c_int]
libapron.ap_texpr1_binop.restype = POINTER(Texpr1)
libapron.ap_texpr1_copy.argtypes = [PyTexpr1]
libapron.ap_texpr1_copy.restype = POINTER(Texpr1)
libapron.ap_texpr1_free.argtypes = [PyTexpr1]
libapron.ap_texpr1_from_linexpr1.argtypes = [PyLinexpr1]
libapron.ap_texpr1_from_linexpr1.restype = POINTER(Texpr1)
libapron.ap_texpr1_substitute.argtypes = [PyTexpr1, c_char_p, PyTexpr1]
libapron.ap_texpr1_substitute.restype = POINTER(Texpr1)



"""
APRON (String) Variables
========================

:Author: Caterina Urban
"""
from _ctypes import Structure
from ctypes import CFUNCTYPE, c_int, c_char_p, c_void_p

from apronpy.cdll import libapron


class VarOperations(Structure):
    """
    typedef struct ap_var_operations_t {
      int (*compare)(ap_var_t v1, ap_var_t v2); /* Total ordering function */
      int (*hash)(ap_var_t v);                  /* Hash function */
      ap_var_t (*copy)(ap_var_t var);           /* Duplication function */
      void (*free)(ap_var_t var);               /* Deallocation function */
      char* (*to_string)(ap_var_t var);         /* Conversion to a dynamically allocated string */
    } ap_var_operations_t;
    """

    _fields_ = [
        ('compare', CFUNCTYPE(c_int, c_void_p, c_void_p)),
        ('hash', CFUNCTYPE(c_int, c_void_p)),
        ('copy', CFUNCTYPE(c_void_p, c_void_p)),
        ('free', CFUNCTYPE(None, c_void_p)),
        ('to_string', CFUNCTYPE(c_char_p, c_char_p))
    ]


class PyVar:

    def __init__(self, name: str):
        self.var = name

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = PyVar(self.var)
        memodict[id(self)] = result
        return result

    @property
    def _as_parameter_(self):
        return self.var.encode('utf-8')

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyVar)
        return argument.var.encode('utf-8')

    def __repr__(self):
        return self.var

    def __lt__(self, other: 'PyVar'):
        assert isinstance(other, PyVar)
        operations = VarOperations.in_dll(libapron, 'ap_var_operations_default')
        return operations.compare(self, other) < 0

    def __le__(self, other: 'PyVar'):
        assert isinstance(other, PyVar)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other):
        assert isinstance(other, PyVar)
        operations = VarOperations.in_dll(libapron, 'ap_var_operations_default')
        return operations.compare(self, other) == 0

    def __ne__(self, other):
        assert isinstance(other, PyVar)
        return not self.__eq__(other)

    def __ge__(self, other):
        assert isinstance(other, PyVar)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other):
        assert isinstance(other, PyVar)
        operations = VarOperations.in_dll(libapron, 'ap_var_operations_default')
        return operations.compare(self, other) > 0



"""
APRON Tree Constraints (Level 0)
================================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER

from apronpy.scalar import Scalar, c_uint, c_size_t
from apronpy.texpr0 import Texpr0


class Tcons0(Structure):
    """
    typedef struct ap_tcons0_t {
      ap_texpr0_t* texpr0;  /* expression */
      ap_constyp_t constyp; /* type of constraint */
      ap_scalar_t* scalar;  /* maybe NULL. For EQMOD constraint, indicates the modulo */
    } ap_tcons0_t;
    """

    _fields_ = [
        ('texpr0', POINTER(Texpr0)),
        ('constyp', c_uint),
        ('scalar', POINTER(Scalar))
    ]


class Tcons0Array(Structure):
    """
    typedef struct ap_tcons0_array_t {
      ap_tcons0_t* p;
      size_t size;
    } ap_tcons0_array_t;
    """

    _fields_ = [
        ('p', POINTER(Tcons0)),
        ('size', c_size_t)
    ]



"""
APRON Linear Expressions (Level 0)
==================================

:Author: Caterina Urban
"""
from _ctypes import Structure, Union, POINTER
from ctypes import c_uint, c_size_t
from enum import IntEnum

from apronpy.cdll import libapron
from apronpy.coeff import Coeff
from apronpy.dimension import Dim


class LinexprDiscr(IntEnum):
    """
    typedef enum ap_linexpr_discr_t {
      AP_LINEXPR_DENSE,
      AP_LINEXPR_SPARSE
    } ap_linexpr_discr_t;
    """
    AP_LINEXPR_DENSE = 0
    AP_LINEXPR_SPARSE = 1


class LinTerm(Structure):
    """
    typedef struct ap_linterm_t {
      ap_dim_t dim;
      ap_coeff_t coeff;
    } ap_linterm_t;
    """

    _fields_ = [
        ('dim', Dim),
        ('coeff', Coeff)
    ]

    def __repr__(self):
        return '{}'.format(self.coeff)


class Linexpr0(Structure):
    """
    typedef struct ap_linexpr0_t {
      ap_coeff_t cst;             /* constant */
      ap_linexpr_discr_t discr;   /* discriminant for array */
      size_t size;             /* size of the array */
      union {
        ap_coeff_t* coeff;     /* array of coefficients */
        ap_linterm_t* linterm; /* array of linear terms */
      } p;
    } ap_linexpr0_t;
    """

    class P(Union):
        """
        union {
          ap_coeff_t* coeff;     /* array of coefficients */
          ap_linterm_t* linterm; /* array of linear terms */
        } p;
        """

        _fields_ = [
            ('coeff', POINTER(Coeff)),
            ('linterm', POINTER(LinTerm))
        ]

    _fields_ = [
        ('cst', Coeff),
        ('discr', c_uint),
        ('size', c_size_t),
        ('p', P)
    ]


libapron.ap_linexpr0_minimize.argtypes = [POINTER(Linexpr0)]
libapron.ap_linexpr0_copy.argtypes = [POINTER(Linexpr0)]
libapron.ap_linexpr0_copy.restype = POINTER(Linexpr0)
libapron.ap_linexpr0_is_integer.argtypes = [POINTER(Linexpr0), c_size_t]
libapron.ap_linexpr0_is_real.argtypes = [POINTER(Linexpr0), c_size_t]
libapron.ap_linexpr0_is_linear.argtypes = [POINTER(Linexpr0)]
libapron.ap_linexpr0_is_quasilinear.argtypes = [POINTER(Linexpr0)]



"""
GMP Multi-Precision Rationals
=============================

:Author: Caterina Urban
"""
from _ctypes import Structure, byref, POINTER
from ctypes import c_int, c_char_p, c_double, c_longlong, c_ulonglong, c_char, create_string_buffer
from typing import Union

from apronpy.cdll import libgmp
from apronpy.mpz import MPZ, MPZ_sizeinbase

MPQ_canonicalize = libgmp.__gmpq_canonicalize
# initialization functions
MPQ_init = libgmp.__gmpq_init
MPQ_clear = libgmp.__gmpq_clear
# assignment functions
MPQ_set = libgmp.__gmpq_set
MPQ_set_si = libgmp.__gmpq_set_si
MPQ_set_d = libgmp.__gmpq_set_d
# conversion functions
MPQ_get_str = libgmp.__gmpq_get_str
# comparison functions
MPQ_cmp = libgmp.__gmpq_cmp  # -1: op1 < op2, 0: op1 == op2, 1: op1 > op2
# arithmetic functions
MPQ_add = libgmp.__gmpq_add
MPQ_sub = libgmp.__gmpq_sub
MPQ_mul = libgmp.__gmpq_mul
MPQ_neg = libgmp.__gmpq_neg
MPQ_abs = libgmp.__gmpq_abs


class MPQ(Structure):
    """
    typedef struct
    {
      __mpz_struct _mp_num;
      __mpz_struct _mp_den;
    } __mpq_struct;
    """
    _fields_ = [
        ('_mp_num', MPZ),
        ('_mp_den', MPZ)
    ]

    def __repr__(self):
        n = MPZ_sizeinbase(self._mp_num, 10)
        d = MPZ_sizeinbase(self._mp_den, 10)
        buffer = create_string_buffer(n + d + 3)
        return MPQ_get_str(buffer, 10, self).decode('utf-8')


class PyMPQ:

    def __init__(self, value_or_numerator: Union[MPQ, float, int] = 0, denominator: int = 1):
        self.mpq = MPQ()
        MPQ_init(self)
        if isinstance(value_or_numerator, MPQ):
            MPQ_set(self, value_or_numerator)
        elif isinstance(value_or_numerator, float):
            MPQ_set_d(self, value_or_numerator)
        else:
            assert isinstance(value_or_numerator, int) and isinstance(denominator, int)
            MPQ_set_si(self, c_longlong(value_or_numerator), c_ulonglong(denominator))
            MPQ_canonicalize(self)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = PyMPQ(self.mpq)
        memodict[id(self)] = result
        return result

    def __del__(self):
        MPQ_clear(self)
        del self.mpq

    @property
    def _as_parameter_(self):
        return byref(self.mpq)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyMPQ)
        return argument

    # noinspection PyProtectedMember
    def __repr__(self):
        n = MPZ_sizeinbase(self.mpq._mp_num, 10)
        d = MPZ_sizeinbase(self.mpq._mp_den, 10)
        buffer = create_string_buffer(n + d + 3)
        return MPQ_get_str(buffer, 10, self).decode('utf-8')

    def __lt__(self, other: 'PyMPQ'):
        assert isinstance(other, PyMPQ)
        return MPQ_cmp(self, other) < 0

    def __le__(self, other: 'PyMPQ'):
        assert isinstance(other, PyMPQ)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: 'PyMPQ'):
        assert isinstance(other, PyMPQ)
        return MPQ_cmp(self, other) == 0

    def __ne__(self, other: 'PyMPQ'):
        assert isinstance(other, PyMPQ)
        return not self.__eq__(other)

    def __ge__(self, other: 'PyMPQ'):
        assert isinstance(other, PyMPQ)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other: 'PyMPQ'):
        assert isinstance(other, PyMPQ)
        return MPQ_cmp(self, other) > 0

    def __add__(self, other: 'PyMPQ') -> 'PyMPQ':
        assert isinstance(other, PyMPQ)
        mpq = type(self)(0, 1)
        MPQ_add(mpq, self, other)
        return mpq

    def __sub__(self, other: 'PyMPQ') -> 'PyMPQ':
        assert isinstance(other, PyMPQ)
        mpq = type(self)(0, 1)
        MPQ_sub(mpq, self, other)
        return mpq

    def __mul__(self, other: 'PyMPQ') -> 'PyMPQ':
        assert isinstance(other, PyMPQ)
        mpq = type(self)(0, 1)
        MPQ_mul(mpq, self, other)
        return mpq

    def __neg__(self) -> 'PyMPQ':
        mpq = type(self)(0, 1)
        MPQ_neg(mpq, self)
        return mpq

    def __abs__(self) -> 'PyMPQ':
        mpq = type(self)(0, 1)
        MPQ_abs(mpq, self)
        return mpq


MPQ_canonicalize.argtypes = [PyMPQ]
# initialization functions
MPQ_init.argtypes = [PyMPQ]
MPQ_clear.argtypes = [PyMPQ]
# assignment functions
MPQ_set.argtypes = [PyMPQ, POINTER(MPQ)]
MPQ_set_si.argtypes = [PyMPQ, c_longlong, c_ulonglong]
MPQ_set_d.argtypes = [PyMPQ, c_double]
# conversion functions
MPQ_get_str.argtypes = [POINTER(c_char), c_int, POINTER(MPQ)]
MPQ_get_str.restype = c_char_p
# comparison functions
MPQ_cmp.argtypes = [PyMPQ, PyMPQ]
# arithmetic functions
MPQ_add.argtypes = [PyMPQ, PyMPQ, PyMPQ]
MPQ_sub.argtypes = [PyMPQ, PyMPQ, PyMPQ]
MPQ_mul.argtypes = [PyMPQ, PyMPQ, PyMPQ]
MPQ_neg.argtypes = [PyMPQ, PyMPQ]
MPQ_abs.argtypes = [PyMPQ, PyMPQ]



"""
APRON Boxes
===========

:Author: Caterina Urban
"""
from _ctypes import POINTER

from apronpy.abstract1 import PyAbstract1
from apronpy.cdll import libboxD, libboxMPQ, libboxMPFR
from apronpy.manager import Manager

libboxD.box_manager_alloc.restype = POINTER(Manager)
libboxMPQ.box_manager_alloc.restype = POINTER(Manager)
libboxMPFR.box_manager_alloc.restype = POINTER(Manager)


class PyBoxD(PyAbstract1):

    manager: POINTER(Manager) = libboxD.box_manager_alloc()


class PyBoxMPQ(PyAbstract1):

    manager: POINTER(Manager) = libboxMPQ.box_manager_alloc()


class PyBoxMPFR(PyAbstract1):

    manager: POINTER(Manager) = libboxMPFR.box_manager_alloc()



"""
APRON Dimensions
================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER
from ctypes import c_uint, c_size_t


class Dim(c_uint):

    def __repr__(self):
        return str(self.value)

    def __lt__(self, other: 'Dim'):
        assert isinstance(other, Dim)
        return self.value < other.value

    def __le__(self, other: 'Dim'):
        assert isinstance(other, Dim)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: 'Dim'):
        assert isinstance(other, Dim)
        return self.value == other.value

    def __ne__(self, other: 'Dim'):
        assert isinstance(other, Dim)
        return not self.__eq__(other)

    def __ge__(self, other: 'Dim'):
        assert isinstance(other, Dim)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other: 'Dim'):
        assert isinstance(other, Dim)
        return self.value > other.value


AP_DIM_MAX = Dim(c_uint(-1).value)


class Dimension(Structure):
    """
    typedef struct ap_dimension_t {
      size_t intdim;
      size_t realdim;
    } ap_dimension_t;
    """

    _fields_ = [
        ('intdim', c_size_t),
        ('realdim', c_size_t)
    ]


class DimChange(Structure):
    """
    typedef struct ap_dimchange_t {
      ap_dim_t* dim;  /* Assumed to be an array of size intdim+realdim */
      size_t intdim ; /* Number of integer dimensions to add/remove */
      size_t realdim; /* Number of real dimensions to add/remove */
    } ap_dimchange_t;
    """

    _fields_ = [
        ('dim', POINTER(Dim)),
        ('intdim', c_size_t),
        ('realdim', c_size_t)
    ]


class DimChange2(Structure):
    """
    typedef struct ap_dimchange2_t {
      ap_dimchange_t* add;    /* If not NULL, specifies the adding new dimensions */
      ap_dimchange_t* remove; /* If not NULL, specifies the removal of dimensions */
    } ap_dimchange2_t;
    """

    _fields_ = [
        ('add', POINTER(DimChange)),
        ('remove', POINTER(DimChange))
    ]


class DimPerm(Structure):
    """
    typedef struct ap_dimperm_t {
      ap_dim_t* dim;    /* Array assumed to be of size size */
      size_t size;
    } ap_dimperm_t;
    """

    _fields_ = [
        ('dim', POINTER(Dim)),
        ('size', c_size_t)
    ]



"""
MPFR Multiprecision Floating-Point Numbers
==========================================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from ctypes import c_long, c_int, c_ulonglong, c_double
from enum import IntEnum
from typing import Union

from apronpy.cdll import libmpfr


# initialization and assignment functions
MPFR_init = libmpfr.mpfr_init
MPFR_clear = libmpfr.mpfr_clear
MPFR_set = libmpfr.mpfr_set
MPFR_set_d = libmpfr.mpfr_set_d
# conversion functions
MPFR_get_d = libmpfr.mpfr_get_d
# comparison functions
MPFR_cmp = libmpfr.mpfr_cmp  # -1: op1 < op2, 0: op1 == op2, 1: op1 > op2
# arithmetic functions
MPFR_add = libmpfr.mpfr_add
MPFR_sub = libmpfr.mpfr_sub
MPFR_mul = libmpfr.mpfr_mul
MPFR_neg = libmpfr.mpfr_neg
MPFR_abs = libmpfr.mpfr_abs


class MPFR(Structure):
    """
    typedef struct {
      mpfr_prec_t  _mpfr_prec;
      mpfr_sign_t  _mpfr_sign;
      mpfr_exp_t   _mpfr_exp;
      mp_limb_t   *_mpfr_d;
    } __mpfr_struct;
    """
    _fields_ = [
        ('_mpfr_prec', c_long),
        ('_mpfr_sign', c_int),
        ('_mpfr_exp', c_long),
        ('_mpfr_d', POINTER(c_ulonglong))
    ]

    def __repr__(self):
        return '{}'.format(MPFR_get_d(self, 0))


class Rnd(IntEnum):
    """
    typedef enum {
      MPFR_RNDN=0,  /* round to nearest, with ties to even */
      MPFR_RNDZ,    /* round toward zero */
      MPFR_RNDU,    /* round toward +Inf */
      MPFR_RNDD,    /* round toward -Inf */
      MPFR_RNDA,    /* round away from zero */
      MPFR_RNDF,    /* faithful rounding */
      MPFR_RNDNA=-1 /* round to nearest, with ties away from zero (mpfr_round) */
    } mpfr_rnd_t;
    """
    MPFR_RNDN = 0
    MPFR_RNDZ = 1
    MPFR_RNDU = 2
    MPFR_RNDD = 3
    MPFR_RNDA = 4
    MPFR_RNDF = 5
    MPFR_RNDNA = -1


class PyMPFR:

    def __init__(self, value: Union[MPFR, int, float], rounding: Rnd = Rnd.MPFR_RNDN):
        self.mpfr = MPFR()
        MPFR_init(self)
        if isinstance(value, MPFR):
            MPFR_set(self, value, rounding)
        else:
            assert isinstance(value, (int, float))
            MPFR_set_d(self, value, rounding)
        self.rounding = rounding

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = PyMPFR(self.mpfr, self.rounding)
        memodict[id(self)] = result
        return result
        
    def __del__(self):
        MPFR_clear(self)
        del self.mpfr
        del self.rounding

    @property
    def _as_parameter_(self):
        return byref(self.mpfr)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyMPFR)
        return argument

    def __repr__(self):
        return '{}'.format(MPFR_get_d(self.mpfr, 0))

    def __lt__(self, other: 'PyMPFR'):
        assert isinstance(other, PyMPFR)
        return MPFR_cmp(self, other) < 0

    def __le__(self, other: 'PyMPFR'):
        assert isinstance(other, PyMPFR)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: 'PyMPFR'):
        assert isinstance(other, PyMPFR)
        return MPFR_cmp(self, other) == 0

    def __ne__(self, other: 'PyMPFR'):
        assert isinstance(other, PyMPFR)
        return not self.__eq__(other)

    def __ge__(self, other: 'PyMPFR'):
        assert isinstance(other, PyMPFR)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other: 'PyMPFR'):
        assert isinstance(other, PyMPFR)
        return MPFR_cmp(self, other) > 0

    def __add__(self, other: 'PyMPFR') -> 'PyMPFR':
        assert isinstance(other, PyMPFR)
        mpfr = type(self)(0)
        MPFR_add(mpfr, self, other)
        return mpfr

    def __sub__(self, other: 'PyMPFR') -> 'PyMPFR':
        assert isinstance(other, PyMPFR)
        mpfr = type(self)(0)
        MPFR_sub(mpfr, self, other)
        return mpfr

    def __mul__(self, other: 'PyMPFR') -> 'PyMPFR':
        assert isinstance(other, PyMPFR)
        mpfr = type(self)(0)
        MPFR_mul(mpfr, self, other)
        return mpfr

    def __neg__(self) -> 'PyMPFR':
        mpfr = type(self)(0)
        MPFR_neg(mpfr, self)
        return mpfr

    def __abs__(self) -> 'PyMPFR':
        mpfr = type(self)(0)
        MPFR_abs(mpfr, self)
        return mpfr


# initialization and assignment functions
MPFR_init.argtypes = [PyMPFR]
MPFR_clear.argtypes = [PyMPFR]
MPFR_set.argtypes = [PyMPFR, POINTER(MPFR), c_int]
MPFR_set_d.argtypes = [PyMPFR, c_double, c_int]
# conversion functions
MPFR_get_d.argtypes = [POINTER(MPFR), c_int]
MPFR_get_d.restype = c_double
# comparison functions
MPFR_cmp.argtypes = [PyMPFR, PyMPFR]
# arithmetic functions
MPFR_add.argtypes = [PyMPFR, PyMPFR, PyMPFR]
MPFR_sub.argtypes = [PyMPFR, PyMPFR, PyMPFR]
MPFR_mul.argtypes = [PyMPFR, PyMPFR, PyMPFR]
MPFR_neg.argtypes = [PyMPFR, PyMPFR]
MPFR_abs.argtypes = [PyMPFR, PyMPFR]



"""
APRON Linear Expressions (Level 1)
==================================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from copy import deepcopy
from ctypes import c_uint, c_size_t, c_char_p

from apronpy.cdll import libapron
from apronpy.coeff import Coeff, CoeffDiscr, PyDoubleIntervalCoeff, PyDoubleScalarCoeff, \
    PyMPQScalarCoeff, PyMPFRScalarCoeff, PyMPQIntervalCoeff, PyMPFRIntervalCoeff, PyCoeff
from apronpy.environment import Environment, PyEnvironment
from apronpy.linexpr0 import Linexpr0, LinexprDiscr
from apronpy.scalar import ScalarDiscr
from apronpy.var import PyVar


class Linexpr1(Structure):
    """
    typedef struct ap_linexpr1_t {
      ap_linexpr0_t* linexpr0;
      ap_environment_t* env;
    } ap_linexpr1_t;
    """

    _fields_ = [
        ('linexpr0', POINTER(Linexpr0)),
        ('env', POINTER(Environment))
    ]

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = libapron.ap_linexpr1_copy(byref(self))
        memodict[id(self)] = result
        return result

    def __repr__(self):
        linexpr0 = self.linexpr0.contents
        env = self.env.contents
        result = ''
        if linexpr0.discr == LinexprDiscr.AP_LINEXPR_DENSE:
            result += ' + '.join(
                '{}·{}'.format(linexpr0.p.coeff[i], env.var_of_dim[i].decode('utf-8'))
                for i in range(linexpr0.size)
            )
            result += ' + {}'.format(linexpr0.cst) if result else '{}'.format(linexpr0.cst)
        else:
            assert linexpr0.discr == LinexprDiscr.AP_LINEXPR_SPARSE
            terms = list()
            for i in range(linexpr0.size):
                coeff = linexpr0.p.linterm[i].coeff
                dim = linexpr0.p.linterm[i].dim.value
                if dim <= linexpr0.size:
                    terms.append('{}·{}'.format(coeff, env.var_of_dim[dim].decode('utf-8')))
            result += ' + '.join(terms)
            result += ' + {}'.format(linexpr0.cst) if result else '{}'.format(linexpr0.cst)
        return result.replace('+ -', '- ')


class PyLinexpr1:

    def __init__(self, linexpr1_or_environment, discr=LinexprDiscr.AP_LINEXPR_SPARSE):
        if isinstance(linexpr1_or_environment, Linexpr1):
            self.linexpr1 = linexpr1_or_environment
        else:
            assert isinstance(linexpr1_or_environment, PyEnvironment)
            size = len(linexpr1_or_environment)
            self.linexpr1 = libapron.ap_linexpr1_make(linexpr1_or_environment, discr, size)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = PyLinexpr1(deepcopy(self.linexpr1))
        memodict[id(self)] = result
        return result

    def __del__(self):
        libapron.ap_linexpr1_clear(self)
        del self.linexpr1

    @property
    def _as_parameter_(self):
        return byref(self.linexpr1)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyLinexpr1)
        return argument

    def __repr__(self):
        return '{}'.format(self.linexpr1)

    def is_integer(self):
        linexpr0 = self.linexpr1.linexpr0
        intdim = self.linexpr1.env.contents.intdim
        return bool(libapron.ap_linexpr0_is_integer(linexpr0, intdim))

    def is_real(self):
        linexpr0 = self.linexpr1.linexpr0
        intdim = self.linexpr1.env.contents.intdim
        return bool(libapron.ap_linexpr0_is_real(linexpr0, intdim))

    def is_linear(self):
        linexpr0 = self.linexpr1.linexpr0
        return bool(libapron.ap_linexpr0_is_linear(linexpr0))

    def is_quasilinear(self):
        linexpr0 = self.linexpr1.linexpr0
        return bool(libapron.ap_linexpr0_is_quasilinear(linexpr0))

    def get_cst(self):
        cst = deepcopy(self.linexpr1.linexpr0.contents.cst)
        if cst.discr == CoeffDiscr.AP_COEFF_INTERVAL:
            if cst.val.interval.contents.inf.contents.discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQIntervalCoeff(cst)
            elif cst.val.interval.contents.inf.contents.discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRIntervalCoeff(cst)
            else:
                assert cst.val.interval.contents.inf.contents.discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleIntervalCoeff(cst)
        else:
            assert cst.discr == CoeffDiscr.AP_COEFF_SCALAR
            if cst.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQScalarCoeff(cst)
            elif cst.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRScalarCoeff(cst)
            else:
                assert cst.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleScalarCoeff(cst)
        return result

    def set_cst(self, cst: PyCoeff):
        libapron.ap_coeff_set(byref(self.linexpr1.linexpr0.contents.cst), cst.coeff)

    def get_coeff(self, var: PyVar):
        coeff = libapron.ap_linexpr1_coeffref(self, var._as_parameter_)
        if coeff.contents.discr == CoeffDiscr.AP_COEFF_INTERVAL:
            discr = coeff.contents.val.interval.contents.inf.contents.discr
            if discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQIntervalCoeff()
            elif discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRIntervalCoeff(0, 0)
            else:  # discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleIntervalCoeff()
        else:  # CoeffDiscr.AP_COEFF_SCALAR
            if coeff.contents.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPQ:
                result = PyMPQScalarCoeff()
            elif coeff.contents.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_MPFR:
                result = PyMPFRScalarCoeff(0)
            else:  # coeff.contents.val.scalar.contents.discr == ScalarDiscr.AP_SCALAR_DOUBLE
                result = PyDoubleScalarCoeff()
        libapron.ap_linexpr1_get_coeff(result, self, var._as_parameter_)
        return result

    def set_coeff(self, var: PyVar, coeff: PyCoeff):
        libapron.ap_coeff_set(libapron.ap_linexpr1_coeffref(self, var._as_parameter_), coeff.coeff)


libapron.ap_linexpr1_make.argtypes = [PyEnvironment, c_uint, c_size_t]
libapron.ap_linexpr1_make.restype = Linexpr1
libapron.ap_linexpr1_copy.argtypes = [POINTER(Linexpr1)]
libapron.ap_linexpr1_copy.restype = Linexpr1
libapron.ap_linexpr1_clear.argtypes = [PyLinexpr1]
libapron.ap_linexpr1_coeffref.argtypes = [PyLinexpr1, c_char_p]
libapron.ap_linexpr1_coeffref.restype = POINTER(Coeff)
libapron.ap_linexpr1_get_coeff.argtypes = [PyCoeff, PyLinexpr1, c_char_p]



"""
APRON Tree Expressions (Level 0)
================================

:Author: Caterina Urban
"""
from _ctypes import Structure, Union, POINTER
from ctypes import c_uint
from enum import IntEnum

from apronpy.cdll import libapron
from apronpy.coeff import Coeff
from apronpy.dimension import Dim
from apronpy.linexpr0 import Linexpr0
from apronpy.mpfr import Rnd


class TexprOp(IntEnum):
    """
    typedef enum ap_texpr_op_t {
      /* Binary operators */
      AP_TEXPR_ADD, AP_TEXPR_SUB, AP_TEXPR_MUL, AP_TEXPR_DIV,
      AP_TEXPR_MOD,  /* either integer or real, no rounding */
      AP_TEXPR_POW,
      /* Unary operators */
      AP_TEXPR_NEG   /* no rounding */,
      AP_TEXPR_CAST, AP_TEXPR_SQRT,
    } ap_texpr_op_t;
    """
    AP_TEXPR_ADD = 0
    AP_TEXPR_SUB = 1
    AP_TEXPR_MUL = 2
    AP_TEXPR_DIV = 3
    AP_TEXPR_MOD = 4
    AP_TEXPR_POW = 5
    AP_TEXPR_NEG = 6
    AP_TEXPR_CAST = 7
    AP_TEXPR_SQRT = 8

    def __repr__(self):
        if self.value == 0:
            return '+'
        elif self.value == 1:
            return '-'
        elif self.value == 2:
            return '·'
        elif self.value == 3:
            return '/'
        elif self.value == 4:
            return '%'
        elif self.value == 5:
            return '^'
        elif self.value == 6:
            return '-'
        elif self.value == 7:
            return 'cast'
        elif self.value == 8:
            return 'sqrt'
        return ValueError('no matching constraint type')


class TexprRtype(IntEnum):
    """
    typedef enum ap_texpr_rtype_t {
      AP_RTYPE_REAL,     // real (no rounding)
      AP_RTYPE_INT,      // integer
      AP_RTYPE_SINGLE,   // IEEE 754 32-bit single precision, e.g.: C's float
      AP_RTYPE_DOUBLE,   // IEEE 754 64-bit double precision, e.g.: C's double
      AP_RTYPE_EXTENDED, // non-standard 80-bit double extended, e.g.: Intel's long double
      AP_RTYPE_QUAD,     // non-standard 128-bit quadruple precision, e.g.: Motorola's long double
      AP_RTYPE_SIZE      // Not to be used !
    } ap_texpr_rtype_t;
    """
    AP_RTYPE_REAL = 0
    AP_RTYPE_INT = 1
    AP_RTYPE_SINGLE = 2
    AP_RTYPE_DOUBLE = 3
    AP_RTYPE_EXTENDED = 4
    AP_RTYPE_QUAD = 5
    AP_RTYPE_SIZE = 6


class TexprRdir(IntEnum):
    """
    typedef enum ap_texpr_rdir_t {
      AP_RDIR_NEAREST = GMP_RNDN, /* Nearest */
      AP_RDIR_ZERO    = GMP_RNDZ, /* Zero (truncation for integers) */
      AP_RDIR_UP      = GMP_RNDU, /* + Infinity */
      AP_RDIR_DOWN    = GMP_RNDD, /* - Infinity */
      AP_RDIR_RND,    /* All possible mode, non deterministically */
      AP_RDIR_SIZE    /* Not to be used ! */
    } ap_texpr_rdir_t;
    """
    AP_RDIR_NEAREST = Rnd.MPFR_RNDN
    AP_RDIR_ZERO = Rnd.MPFR_RNDZ
    AP_RDIR_UP = Rnd.MPFR_RNDU
    AP_RDIR_DOWN = Rnd.MPFR_RNDD
    AP_RDIR_RND = 4
    AP_RDIR_SIZE = 5


class Texpr0Node(Structure):
    """
    typedef struct ap_texpr0_node_t {
      ap_texpr_op_t    op;
      ap_texpr_rtype_t type;
      ap_texpr_rdir_t  dir;
      struct ap_texpr0_t* exprA; /* First operand */
      struct ap_texpr0_t* exprB; /* Second operand (for binary operations) or NULL */
    } ap_texpr0_node_t;
    """
    pass


class TexprDiscr(IntEnum):
    """
    typedef enum ap_texpr_discr_t {
      AP_TEXPR_CST, AP_TEXPR_DIM, AP_TEXPR_NODE
    } ap_texpr_discr_t;
    """
    AP_TEXPR_CST = 0
    AP_TEXPR_DIM = 1
    AP_TEXPR_NODE = 2


class Texpr0(Structure):
    """
    typedef struct ap_texpr0_t {
      ap_texpr_discr_t discr;
      union {
        ap_coeff_t cst;
        ap_dim_t dim;
        ap_texpr0_node_t* node;
      } val;
    } ap_texpr0_t;
    """

    class Val(Union):
        """
        union {
          ap_coeff_t cst;
          ap_dim_t dim;
          ap_texpr0_node_t* node;
        } val;
        """

        _fields_ = [
            ('cst', Coeff),
            ('dim', Dim),
            ('node', POINTER(Texpr0Node))
        ]

    _fields_ = [
        ('discr', c_uint),
        ('val', Val)
    ]


Texpr0Node._fields_ = [
        ('op', c_uint),
        ('type', c_uint),
        ('dir', c_uint),
        ('exprA', POINTER(Texpr0)),
        ('exprB', POINTER(Texpr0))
    ]


libapron.ap_texpr0_copy.argtypes = [POINTER(Texpr0)]
libapron.ap_texpr0_copy.restype = POINTER(Texpr0)
libapron.ap_texpr0_free.argtypes = [POINTER(Texpr0)]
libapron.ap_texpr0_from_linexpr0.argtypes = [POINTER(Linexpr0)]
libapron.ap_texpr0_from_linexpr0.restype = POINTER(Texpr0)
libapron.ap_texpr0_substitute.argtypes = [POINTER(Texpr0), Dim, POINTER(Texpr0)]
libapron.ap_texpr0_substitute.restype = POINTER(Texpr0)
libapron.ap_texpr0_substitute_with.argtypes = [POINTER(Texpr0), Dim, POINTER(Texpr0)]



"""
APRON Tree Constraints (Level 1)
================================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from copy import deepcopy
from ctypes import c_size_t
from typing import List, Union

from apronpy.cdll import libapron
from apronpy.coeff import PyDoubleScalarCoeff
from apronpy.environment import Environment, PyEnvironment
from apronpy.lincons0 import ConsTyp
from apronpy.lincons1 import PyLincons1
from apronpy.scalar import c_uint, PyScalar
from apronpy.tcons0 import Tcons0, Tcons0Array
from apronpy.texpr0 import TexprOp, TexprDiscr
from apronpy.texpr1 import PyTexpr1
from apronpy.var import PyVar


class Tcons1(Structure):
    """
    typedef struct ap_tcons1_t {
      ap_tcons0_t tcons0;
      ap_environment_t* env;
    } ap_tcons1_t;
    """

    _fields_ = [
        ('tcons0', Tcons0),
        ('env', POINTER(Environment))
    ]

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = Tcons1()
        result.tcons0 = Tcons0()
        result.tcons0.texpr0 = libapron.ap_texpr0_copy(self.tcons0.texpr0)
        result.tcons0.constyp = self.tcons0.constyp
        if self.tcons0.scalar:
            result.tcons0.scalar = libapron.ap_scalar_alloc_set(self.tcons0.scalar)
        else:
            result.tcons0.scalar = None
        self.env.contents.count += 1
        result.env = self.env
        memodict[id(self)] = result
        return result

    def __repr__(self):
        def precendence(texpr0):
            op_precedence = {
                TexprOp.AP_TEXPR_ADD: 1,
                TexprOp.AP_TEXPR_SUB: 1,
                TexprOp.AP_TEXPR_MUL: 2,
                TexprOp.AP_TEXPR_DIV: 2,
                TexprOp.AP_TEXPR_MOD: 2,
                TexprOp.AP_TEXPR_POW: 3,
                TexprOp.AP_TEXPR_NEG: 4,
                TexprOp.AP_TEXPR_CAST: 5,
                TexprOp.AP_TEXPR_SQRT: 5
            }
            if texpr0.discr == TexprDiscr.AP_TEXPR_CST or texpr0.discr == TexprDiscr.AP_TEXPR_DIM:
                return op_precedence[TexprOp.AP_TEXPR_NEG]
            return op_precedence[texpr0.val.node.contents.op]

        def do(texpr0, env):
            if texpr0.discr == TexprDiscr.AP_TEXPR_CST:
                return '{}'.format(texpr0.val.cst)
            elif texpr0.discr == TexprDiscr.AP_TEXPR_DIM:
                return '{}'.format(env.var_of_dim[texpr0.val.dim.value].decode('utf-8'))
            else:  # texpr0.discr == TexprDiscr.AP_TEXPR_NODE
                prec = precendence(texpr0)
                prec_a = precendence(texpr0.val.node.contents.exprA.contents)
                if prec_a < prec:
                    expr_a = '(' + do(texpr0.val.node.contents.exprA.contents, env) + ')'
                else:
                    expr_a = do(texpr0.val.node.contents.exprA.contents, env)
                op = texpr0.val.node.contents.op
                if texpr0.val.node.contents.exprB:  # binary operation
                    prec_b = precendence(texpr0.val.node.contents.exprB.contents)
                    if prec_b <= prec:
                        expr_b = '(' + do(texpr0.val.node.contents.exprB.contents, env) + ')'
                    else:
                        expr_b = do(texpr0.val.node.contents.exprB.contents, env)
                    return '{} {} {}'.format(expr_a, repr(TexprOp(op)), expr_b)
                else:
                    return '{} {}'.format(repr(TexprOp(op)), expr_a)

        constyp = ConsTyp(self.tcons0.constyp)
        scalar = self.tcons0.scalar
        result = do(self.tcons0.texpr0.contents, self.env.contents).replace('+ -', '- ')
        if scalar:
            return '{} {} {}'.format(result, repr(constyp), scalar.contents)
        else:
            return '{} {} 0'.format(result, repr(constyp))


class TCons1Array(Structure):
    """
    typedef struct ap_tcons1_array_t {
      ap_tcons0_array_t tcons0_array;
      ap_environment_t* env;
    } ap_tcons1_array_t;
    """

    _fields_ = [
        ('tcons0_array', Tcons0Array),
        ('env', POINTER(Environment))
    ]

    def __repr__(self):
        def precendence(texpr0):
            op_precedence = {
                TexprOp.AP_TEXPR_ADD: 1,
                TexprOp.AP_TEXPR_SUB: 1,
                TexprOp.AP_TEXPR_MUL: 2,
                TexprOp.AP_TEXPR_DIV: 2,
                TexprOp.AP_TEXPR_MOD: 2,
                TexprOp.AP_TEXPR_POW: 3,
                TexprOp.AP_TEXPR_NEG: 4,
                TexprOp.AP_TEXPR_CAST: 5,
                TexprOp.AP_TEXPR_SQRT: 5
            }
            if texpr0.discr == TexprDiscr.AP_TEXPR_CST or texpr0.discr == TexprDiscr.AP_TEXPR_DIM:
                return op_precedence[TexprOp.AP_TEXPR_NEG]
            return op_precedence[texpr0.val.node.contents.op]

        def do(texpr0, env):
            if texpr0.discr == TexprDiscr.AP_TEXPR_CST:
                return '{}'.format(texpr0.val.cst)
            elif texpr0.discr == TexprDiscr.AP_TEXPR_DIM:
                return '{}'.format(env.var_of_dim[texpr0.val.dim.value].decode('utf-8'))
            else:  # texpr0.discr == TexprDiscr.AP_TEXPR_NODE
                prec = precendence(texpr0)
                prec_a = precendence(texpr0.val.node.contents.exprA.contents)
                if prec_a < prec:
                    expr_a = '(' + do(texpr0.val.node.contents.exprA.contents, env) + ')'
                else:
                    expr_a = do(texpr0.val.node.contents.exprA.contents, env)
                op = texpr0.val.node.contents.op
                if texpr0.val.node.contents.exprB:  # binary operation
                    prec_b = precendence(texpr0.val.node.contents.exprB.contents)
                    if prec_b <= prec:
                        expr_b = '(' + do(texpr0.val.node.contents.exprB.contents, env) + ')'
                    else:
                        expr_b = do(texpr0.val.node.contents.exprB.contents, env)
                    return '{} {} {}'.format(expr_a, repr(TexprOp(op)), expr_b)
                else:
                    return '{} {}'.format(repr(TexprOp(op)), expr_a)

        array = list()
        for i in range(self.tcons0_array.size):
            constyp = ConsTyp(self.tcons0_array.p[i].constyp)
            scalar = self.tcons0_array.p[i].scalar
            result = do(self.tcons0_array.p[i].texpr0.contents, self.env.contents)
            result = result.replace('+ -', '- ')
            if scalar:
                array.append('{} {} {}'.format(result, repr(constyp), scalar.contents))
            else:
                array.append('{} {} 0'.format(result, repr(constyp)))
        return ' ∧ '.join(array)


class PyTcons1:

    def __init__(self, cons: Union[PyLincons1, Tcons1]):
        if isinstance(cons, PyLincons1):
            self.tcons1 = Tcons1()
            texpr = libapron.ap_texpr0_from_linexpr0(cons.lincons1.lincons0.linexpr0)
            self.tcons1.tcons0.texpr0 = texpr
            self.tcons1.tcons0.constyp = c_uint(cons.lincons1.lincons0.constyp)
            scalar = cons.lincons1.lincons0.scalar
            if scalar:
                self.tcons1.tcons0.scalar = libapron.ap_scalar_alloc_set(scalar)
            cons.lincons1.env.contents.count += 1
            self.tcons1.env = cons.lincons1.env
        else:
            self.tcons1 = cons

    @classmethod
    def make(cls, texpr: PyTexpr1, typ: ConsTyp, cst: PyScalar = None):
        tcons1 = Tcons1()
        tcons1.tcons0 = Tcons0()
        tcons1.tcons0.texpr0 = libapron.ap_texpr0_copy(texpr.texpr1.contents.texpr0)
        tcons1.tcons0.constyp = c_uint(typ)
        if cst:
            tcons1.tcons0.scalar = libapron.ap_scalar_alloc_set(cst)
        else:
            tcons1.tcons0.scalar = None
        texpr.texpr1.contents.env.contents.count += 1
        tcons1.env = texpr.texpr1.contents.env
        return cls(tcons1)

    @classmethod
    def unsat(cls, environment: PyEnvironment):
        x = PyTexpr1.cst(environment, PyDoubleScalarCoeff(-1.0))
        return cls.make(x, ConsTyp.AP_CONS_SUPEQ)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = PyTcons1(deepcopy(self.tcons1))
        memodict[id(self)] = result
        return result

    def __del__(self):
        libapron.ap_tcons1_clear(self)
        del self.tcons1

    @property
    def _as_parameter_(self):
        return byref(self.tcons1)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyTcons1)
        return argument

    def __repr__(self):
        return '{}'.format(self.tcons1)

    def substitute(self, var: PyVar, dst: PyTexpr1):
        self.tcons1.env.contents.count += 1
        dim = libapron.ap_environment_dim_of_var(PyEnvironment(self.tcons1.env), var)
        texpr0 = self.tcons1.tcons0.texpr0
        libapron.ap_texpr0_substitute_with(texpr0, dim, dst.texpr1.contents.texpr0)
        return self


libapron.ap_tcons1_clear.argtypes = [PyTcons1]


class PyTcons1Array:

    def __init__(self, tcons1s: Union[TCons1Array, List[PyTcons1]] = None,
                 environment: PyEnvironment = None):
        if isinstance(tcons1s, TCons1Array):
            self.tcons1array = tcons1s
        elif tcons1s:
            size = len(tcons1s)
            self.tcons1array = libapron.ap_tcons1_array_make(tcons1s[0].tcons1.env, size)
            for i in range(size):
                libapron.ap_tcons1_array_set(self, i, deepcopy(tcons1s[i].tcons1))
        else:
            self.tcons1array = libapron.ap_tcons1_array_make(environment, 0)

    def __del__(self):
        libapron.ap_tcons1_array_clear(self)
        del self.tcons1array

    @property
    def _as_parameter_(self):
        return byref(self.tcons1array)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyTcons1Array)
        return argument

    def __repr__(self):
        return '{}'.format(self.tcons1array)


libapron.ap_tcons1_array_make.argtypes = [POINTER(Environment), c_size_t]
libapron.ap_tcons1_array_make.restype = TCons1Array
libapron.ap_tcons1_array_clear.argtypes = [PyTcons1Array]
libapron.ap_tcons1_array_set.argtypes = [PyTcons1Array, c_size_t, POINTER(Tcons1)]



"""
APRON Intervals on Scalars
==========================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from abc import ABCMeta
from ctypes import c_double, c_int

from apronpy.cdll import libapron
from apronpy.mpfr import PyMPFR, Rnd
from apronpy.mpq import PyMPQ
from apronpy.scalar import Scalar, PyScalar, PyDoubleScalar, PyMPQScalar, PyMPFRScalar


class Interval(Structure):
    """
    typedef struct ap_interval_t {
      ap_scalar_t* inf;
      ap_scalar_t* sup;
    } ap_interval_t;
    """

    _fields_ = [
        ('inf', POINTER(Scalar)),
        ('sup', POINTER(Scalar))
    ]

    def __repr__(self):
        return '[{},{}]'.format(self.inf.contents, self.sup.contents)


class PyInterval(metaclass=ABCMeta):

    def __init__(self, value_or_inf, sup=None):
        self.interval = libapron.ap_interval_alloc()
        if isinstance(value_or_inf, Interval):
            libapron.ap_interval_set(self, byref(value_or_inf))
        elif isinstance(value_or_inf, c_double) and isinstance(sup, c_double):
            libapron.ap_interval_set_double(self, value_or_inf, sup)
        elif isinstance(value_or_inf, PyMPQ) and isinstance(sup, PyMPQ):
            libapron.ap_interval_set_mpq(self, value_or_inf, sup)
        elif isinstance(value_or_inf, PyMPFR) and isinstance(sup, PyMPFR):
            libapron.ap_interval_set_mpfr(self, value_or_inf, sup)
        else:
            assert isinstance(value_or_inf, PyScalar) and isinstance(sup, PyScalar)
            libapron.ap_interval_set_scalar(self, value_or_inf, sup)

    @classmethod
    def top(cls):
        interval = cls(0, 0)
        libapron.ap_interval_set_top(interval)
        return interval

    @classmethod
    def bottom(cls):
        interval = cls(0, 0)
        libapron.ap_interval_set_bottom(interval)
        return interval

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        result = type(self)(self.interval.contents)
        memodict[id(self)] = result
        return result

    def __del__(self):
        libapron.ap_interval_free(self)
        del self.interval

    @property
    def _as_parameter_(self):
        return self.interval

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyInterval)
        return argument

    def __repr__(self):
        return str(self.interval.contents)

    def is_bottom(self):
        return bool(libapron.ap_interval_is_bottom(self))
    
    def __lt__(self, other: 'PyInterval'):
        assert isinstance(other, PyInterval)
        return libapron.ap_interval_cmp(self, other) == -1

    def __le__(self, other: 'PyInterval'):
        assert isinstance(other, PyInterval)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other):
        assert isinstance(other, PyInterval)
        return libapron.ap_interval_cmp(self, other) == 0

    def __ne__(self, other):
        assert isinstance(other, PyInterval)
        return not self.__eq__(other)

    def __ge__(self, other):
        assert isinstance(other, PyInterval)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other):
        assert isinstance(other, PyInterval)
        return libapron.ap_interval_cmp(self, other) == 1

    def is_top(self):
        return bool(libapron.ap_interval_is_top(self))

    def __neg__(self) -> 'PyInterval':
        interval = type(self)(0, 0)
        libapron.ap_interval_neg(interval, self)
        return interval


libapron.ap_interval_alloc.restype = POINTER(Interval)
libapron.ap_interval_set.argtypes = [PyInterval, POINTER(Interval)]
libapron.ap_interval_set_int.argtypes = [PyInterval, c_int, c_int]
libapron.ap_interval_set_double.argtypes = [PyInterval, c_double, c_double]
libapron.ap_interval_set_mpq.argtypes = [PyInterval, PyMPQ, PyMPQ]
libapron.ap_interval_set_mpfr.argtypes = [PyInterval, PyMPFR, PyMPFR]
libapron.ap_interval_set_scalar.argtypes = [PyInterval, PyScalar, PyScalar]
libapron.ap_interval_is_bottom.argtypes = [PyInterval]
libapron.ap_interval_is_top.argtypes = [PyInterval]
libapron.ap_interval_cmp.argtypes = [PyInterval, PyInterval]
libapron.ap_interval_neg.argtypes = [PyInterval, PyInterval]


class PyDoubleInterval(PyInterval):

    def __init__(self, value_or_inf=0.0, sup=0.0):
        if isinstance(value_or_inf, Interval):
            super().__init__(value_or_inf)
        elif isinstance(value_or_inf, PyDoubleScalar) and isinstance(sup, PyDoubleScalar):
            super().__init__(value_or_inf, sup)
        elif isinstance(value_or_inf, c_double) and isinstance(sup, c_double):
            super().__init__(value_or_inf, sup)
        else:
            super().__init__(c_double(value_or_inf), c_double(sup))


class PyMPQInterval(PyInterval):

    def __init__(self, value_or_inf_num=0, sup_num=0, inf_den=1, sup_den=1):
        if isinstance(value_or_inf_num, Interval):
            super().__init__(value_or_inf_num)
        elif isinstance(value_or_inf_num, PyMPQScalar) and isinstance(sup_num, PyMPQScalar):
            super().__init__(value_or_inf_num, sup_num)
        elif isinstance(value_or_inf_num, PyMPQ) and isinstance(sup_num, PyMPQ):
            super().__init__(value_or_inf_num, sup_num)
        else:
            super().__init__(PyMPQ(value_or_inf_num, inf_den), PyMPQ(sup_num, sup_den))


class PyMPFRInterval(PyInterval):

    def __init__(self, value_or_inf, sup=None, rounding: Rnd = Rnd.MPFR_RNDN):
        if isinstance(value_or_inf, Interval):
            super().__init__(value_or_inf)
        elif isinstance(value_or_inf, PyMPFRScalar) and isinstance(sup, PyMPFRScalar):
            super().__init__(value_or_inf, sup)
        elif isinstance(value_or_inf, PyMPFR) and isinstance(sup, PyMPFR):
            super().__init__(value_or_inf, sup)
        else:
            super().__init__(PyMPFR(value_or_inf, rounding), PyMPFR(sup, rounding))



"""
APRON Abstract Values (Level 1)
===============================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from abc import ABCMeta
from ctypes import c_size_t, c_char_p, c_bool
from typing import List, Type, Union

from apronpy.lincons1 import Lincons1Array, PyLincons1Array
from apronpy.linexpr1 import PyLinexpr1, Linexpr1

from apronpy.abstract0 import Abstract0
from apronpy.cdll import libapron
from apronpy.environment import Environment, PyEnvironment
from apronpy.interval import Interval, PyInterval
from apronpy.manager import Manager
from apronpy.tcons1 import PyTcons1Array, TCons1Array
from apronpy.texpr1 import PyTexpr1, Texpr1
from apronpy.var import PyVar

APRON_substitute_linexpr_array = libapron.ap_abstract1_substitute_linexpr_array
APRON_substitute_texpr_array = libapron.ap_abstract1_substitute_texpr_array


class Abstract1(Structure):
    """
    typedef struct ap_abstract1_t {
      ap_abstract0_t* abstract0;
      ap_environment_t* env;
    } ap_abstract1_t;
    """

    _fields_ = [
        ('abstract0', POINTER(Abstract0)),
        ('env', POINTER(Environment))
    ]


class PyAbstract1(metaclass=ABCMeta):

    manager: POINTER(Manager)

    # noinspection PyTypeChecker
    def __init__(self, abstract1_or_environment: Union[Abstract1, PyEnvironment],
                 bottom: bool = False, array: Union[PyLincons1Array, PyTcons1Array] = None,
                 variables: List[PyVar] = None, intervals: List[PyInterval] = None):
        if isinstance(abstract1_or_environment, Abstract1):
            self.abstract1 = abstract1_or_environment
        elif bottom:
            assert isinstance(abstract1_or_environment, PyEnvironment)
            self.abstract1 = libapron.ap_abstract1_bottom(self.manager, abstract1_or_environment)
        elif array and isinstance(array, PyLincons1Array):
            assert isinstance(abstract1_or_environment, PyEnvironment)
            man = self.manager
            a1 = libapron.ap_abstract1_of_lincons_array(man, abstract1_or_environment, array)
            self.abstract1 = a1
        elif array and isinstance(array, PyTcons1Array):
            assert isinstance(abstract1_or_environment, PyEnvironment)
            man = self.manager
            a1 = libapron.ap_abstract1_of_tcons_array(man, abstract1_or_environment, array)
            self.abstract1 = a1
        elif variables and intervals:
            assert isinstance(abstract1_or_environment, PyEnvironment)
            size = len(variables)
            v_typ: Type = c_char_p * size
            v_arr = v_typ(*(x._as_parameter_ for x in variables))
            i_typ: Type = POINTER(Interval) * size
            i_arr = i_typ(*(x._as_parameter_ for x in intervals))
            man = self.manager
            a1 = libapron.ap_abstract1_of_box(man, abstract1_or_environment, v_arr, i_arr, size)
            self.abstract1 = a1
        else:
            assert isinstance(abstract1_or_environment, PyEnvironment)
            self.abstract1 = libapron.ap_abstract1_top(self.manager, abstract1_or_environment)

    @classmethod
    def bottom(cls, environment: PyEnvironment):
        return cls(environment, bottom=True)

    @classmethod
    def from_constraints(cls, environment: PyEnvironment, array):
        return cls(environment, array=array)

    @classmethod
    def top(cls, environment: PyEnvironment):
        return cls(environment)

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        abstract1 = libapron.ap_abstract1_copy(self.manager, self)
        result = type(self)(abstract1)
        memodict[id(self)] = result
        return result

    def __del__(self):
        libapron.ap_abstract1_clear(self.manager, self)
        del self.abstract1

    @property
    def _as_parameter_(self):
        return byref(self.abstract1)

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyAbstract1)
        return argument

    @property
    def to_lincons(self) -> PyLincons1Array:
        return PyLincons1Array(libapron.ap_abstract1_to_lincons_array(self.manager, self))

    @property
    def to_tcons(self) -> PyTcons1Array:
        return PyTcons1Array(libapron.ap_abstract1_to_tcons_array(self.manager, self))

    def __repr__(self):
        array = PyLincons1Array(libapron.ap_abstract1_to_lincons_array(self.manager, self))
        return '{}'.format(array)

    @property
    def environment(self) -> PyEnvironment:
        return PyEnvironment(libapron.ap_abstract1_environment(self.manager, self))

    def is_bottom(self):
        return bool(libapron.ap_abstract1_is_bottom(self.manager, self))

    def is_top(self):
        return bool(libapron.ap_abstract1_is_top(self.manager, self))

    def __le__(self, other: 'PyAbstract1'):
        assert isinstance(other, PyAbstract1)
        return bool(libapron.ap_abstract1_is_leq(self.manager, self, other))

    def is_leq(self, other: 'PyAbstract1'):
        assert isinstance(other, PyAbstract1)
        return self.__le__(other)

    def __eq__(self, other: 'PyAbstract1'):
        assert isinstance(other, PyAbstract1)
        return bool(libapron.ap_abstract1_is_eq(self.manager, self, other))

    def is_eq(self, other: 'PyAbstract1'):
        assert isinstance(other, PyAbstract1)
        return self.__eq__(other)

    def meet(self, other: Union['PyAbstract1', PyLincons1Array, PyTcons1Array]):
        if isinstance(other, PyLincons1Array):
            abstract1 = libapron.ap_abstract1_meet_lincons_array(self.manager, False, self, other)
            return type(self)(abstract1)
        elif isinstance(other, PyTcons1Array):
            abstract1 = libapron.ap_abstract1_meet_tcons_array(self.manager, False, self, other)
            return type(self)(abstract1)
        else:
            assert isinstance(other, PyAbstract1)
            abstract1 = libapron.ap_abstract1_meet(self.manager, False, self, other)
            return type(self)(abstract1)

    def join(self, other: 'PyAbstract1'):
        assert isinstance(other, PyAbstract1)
        return type(self)(libapron.ap_abstract1_join(self.manager, False, self, other))

    def widening(self, other: 'PyAbstract1'):
        assert isinstance(other, PyAbstract1)
        return type(self)(libapron.ap_abstract1_widening(self.manager, self, other))

    def assign(self, var: PyVar, expr: Union[PyLinexpr1, PyTexpr1]):
        man = self.manager
        if isinstance(expr, PyLinexpr1):
            abstract1 = libapron.ap_abstract1_assign_linexpr(man, False, self, var, expr, None)
            return type(self)(abstract1)
        else:
            assert isinstance(expr, PyTexpr1)
            abstract1 = libapron.ap_abstract1_assign_texpr(man, False, self, var, expr, None)
            return type(self)(abstract1)

    # noinspection PyTypeChecker
    def substitute(self, var_or_vars: Union[PyVar, List[PyVar]],
                   expr_or_exprs: Union[PyLinexpr1, PyTexpr1, List[PyLinexpr1], List[PyTexpr1]]):
        man = self.manager
        if isinstance(var_or_vars, PyVar):
            var = var_or_vars
            expr = expr_or_exprs
            if isinstance(expr, PyLinexpr1):
                a1 = libapron.ap_abstract1_substitute_linexpr(man, False, self, var, expr, None)
                return type(self)(a1)
            else:
                assert isinstance(expr, PyTexpr1)
                a1 = libapron.ap_abstract1_substitute_texpr(man, False, self, var, expr, None)
                return type(self)(a1)
        else:
            assert isinstance(var_or_vars, list)
            assert all(isinstance(var, PyVar) for var in var_or_vars)
            exprs = expr_or_exprs
            if all(isinstance(expr, PyLinexpr1) for expr in exprs):
                v_size = len(var_or_vars)
                e_size = len(exprs)
                assert v_size == e_size
                v_typ: Type = c_char_p * v_size
                v_arr = v_typ(*(x._as_parameter_ for x in var_or_vars))
                e_typ: Type = Linexpr1 * e_size
                e_arr = e_typ(*(e.linexpr1 for e in exprs))
                a1 = APRON_substitute_linexpr_array(man, False, self, v_arr, e_arr, v_size, None)
                return type(self)(a1)
            else:
                assert all(isinstance(expr, PyTexpr1) for expr in exprs)
                v_size = len(var_or_vars)
                e_size = len(exprs)
                assert v_size == e_size
                v_typ: Type = c_char_p * v_size
                v_arr = v_typ(*(x._as_parameter_ for x in var_or_vars))
                e_typ: Type = Texpr1 * e_size
                e_arr = e_typ(*(e.texpr1.contents for e in exprs))
                a1 = APRON_substitute_texpr_array(man, False, self, v_arr, e_arr, v_size, None)
                return type(self)(a1)


man_p = POINTER(Manager)
pya1 = PyAbstract1
libapron.ap_abstract1_copy.argtypes = [man_p, pya1]
libapron.ap_abstract1_copy.restype = Abstract1
libapron.ap_abstract1_clear.argtypes = [man_p, pya1]
libapron.ap_abstract1_bottom.argtypes = [man_p, PyEnvironment]
libapron.ap_abstract1_bottom.restype = Abstract1
pyvar_p = POINTER(c_char_p)
pyitv_p = POINTER(POINTER(Interval))
libapron.ap_abstract1_of_box.argtypes = [man_p, PyEnvironment, pyvar_p, pyitv_p, c_size_t]
libapron.ap_abstract1_of_box.restype = Abstract1
libapron.ap_abstract1_top.argtypes = [man_p, PyEnvironment]
libapron.ap_abstract1_top.restype = Abstract1
libapron.ap_abstract1_environment.argtypes = [man_p, pya1]
libapron.ap_abstract1_environment.restype = POINTER(Environment)
libapron.ap_abstract1_is_bottom.argtypes = [man_p, pya1]
libapron.ap_abstract1_is_top.argtypes = [man_p, pya1]
libapron.ap_abstract1_is_leq.argtypes = [man_p, pya1, pya1]
libapron.ap_abstract1_is_eq.argtypes = [man_p, pya1, pya1]
libapron.ap_abstract1_to_lincons_array.argtypes = [man_p, pya1]
libapron.ap_abstract1_to_lincons_array.restype = Lincons1Array
libapron.ap_abstract1_to_tcons_array.argtypes = [man_p, pya1]
libapron.ap_abstract1_to_tcons_array.restype = TCons1Array
libapron.ap_abstract1_meet.argtypes = [man_p, c_bool, pya1, pya1]
libapron.ap_abstract1_meet.restype = Abstract1
libapron.ap_abstract1_meet_lincons_array.argtypes = [man_p, c_bool, pya1, PyLincons1Array]
libapron.ap_abstract1_meet_lincons_array.restype = Abstract1
libapron.ap_abstract1_meet_tcons_array.argtypes = [man_p, c_bool, pya1, PyTcons1Array]
libapron.ap_abstract1_meet_tcons_array.restype = Abstract1
libapron.ap_abstract1_join.argtypes = [man_p, c_bool, pya1, pya1]
libapron.ap_abstract1_join.restype = Abstract1
libapron.ap_abstract1_widening.argtypes = [man_p, pya1, pya1]
libapron.ap_abstract1_widening.restype = Abstract1
pyl1 = PyLinexpr1
pya1_p = POINTER(Abstract1)
libapron.ap_abstract1_of_lincons_array.argtypes = [man_p, PyEnvironment, PyLincons1Array]
libapron.ap_abstract1_of_lincons_array.restype = Abstract1
libapron.ap_abstract1_of_tcons_array.argtypes = [man_p, PyEnvironment, PyTcons1Array]
libapron.ap_abstract1_of_tcons_array.restype = Abstract1
libapron.ap_abstract1_assign_linexpr.argtypes = [man_p, c_bool, pya1, PyVar, pyl1, pya1_p]
libapron.ap_abstract1_assign_linexpr.restype = Abstract1
pyl1_p = POINTER(Linexpr1)
APRON_substitute_linexpr_array.argtypes = [man_p, c_bool, pya1, pyvar_p, pyl1_p, c_size_t, pya1_p]
APRON_substitute_linexpr_array.restype = Abstract1
libapron.ap_abstract1_substitute_linexpr.argtypes = [man_p, c_bool, pya1, PyVar, pyl1, pya1_p]
libapron.ap_abstract1_substitute_linexpr.restype = Abstract1
libapron.ap_abstract1_assign_texpr.argtypes = [man_p, c_bool, pya1, PyVar, PyTexpr1, pya1_p]
libapron.ap_abstract1_assign_texpr.restype = Abstract1
pyt1_p = POINTER(Texpr1)
APRON_substitute_texpr_array.argtypes = [man_p, c_bool, pya1, pyvar_p, pyt1_p, c_size_t, pya1_p]
APRON_substitute_texpr_array.restype = Abstract1
libapron.ap_abstract1_substitute_texpr.argtypes = [man_p, c_bool, pya1, PyVar, PyTexpr1, pya1_p]
libapron.ap_abstract1_substitute_texpr.restype = Abstract1



"""
APRON Environments
==================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER, byref
from ctypes import c_size_t, c_char_p
from typing import List, Type, Union

from apronpy.cdll import libapron
from apronpy.dimension import Dim, AP_DIM_MAX, DimChange, DimPerm
from apronpy.var import PyVar


class Environment(Structure):
    """
    typedef struct ap_environment_t {
      ap_var_t* var_of_dim;
      /*
        Array of size intdim+realdim, indexed by dimensions.
        - It should not contain identical strings..
        - Slice [0..intdim-1] is lexicographically sorted,
          and denotes integer variables.
        - Slice [intdim..intdim+realdim-1] is lexicographically sorted,
          and denotes real variables.
        - The memory allocated for the variables are attached to the structure
          (they are freed when the structure is no longer in use)
      */
      size_t intdim; /* Number of integer variables */
      size_t realdim;/* Number of real variables */
      size_t count; /* For reference counting */
    } ap_environment_t;
    """

    _fields_ = [
        ('var_of_dim', POINTER(c_char_p)),
        ('intdim', c_size_t),
        ('realdim', c_size_t),
        ('count', c_size_t)
    ]

    def __repr__(self):
        result = '{'
        result += ','.join(self.var_of_dim[i].decode('utf-8') for i in range(self.intdim))
        result += '|'
        result += ','.join(
            self.var_of_dim[self.intdim+i].decode('utf-8') for i in range(self.realdim)
        )
        result += '}'
        return result


class PyEnvironment:

    # noinspection PyTypeChecker
    def __init__(self, environment_or_int_vars: Union[POINTER(Environment), List[PyVar]] = None,
                 real_vars: List[PyVar] = None):
        if isinstance(environment_or_int_vars, POINTER(Environment)):
            self.environment = environment_or_int_vars
        else:
            if environment_or_int_vars:
                int_size = len(environment_or_int_vars)
                typ: Type = c_char_p * int_size
                int_arr = typ(*(x._as_parameter_ for x in environment_or_int_vars))
            else:
                int_size = 0
                int_arr = None
            if real_vars:
                real_size = len(real_vars)
                typ: Type = c_char_p * real_size
                real_arr = typ(*(x._as_parameter_ for x in real_vars))
            else:
                real_size = 0
                real_arr = None
            environment = libapron.ap_environment_alloc(int_arr, int_size, real_arr, real_size)
            self.environment = environment
            if not self.environment:
                raise ValueError('clashing variable names')

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        self.environment.contents.count += 1
        memodict[id(self)] = self
        return self

    def __del__(self):
        if self.environment:
            if self.environment.contents.count <= 1:
                libapron.ap_environment_free2(self)
                del self.environment
            else:
                self.environment.contents.count -= 1

    @property
    def _as_parameter_(self):
        return self.environment

    @staticmethod
    def from_param(argument):
        assert isinstance(argument, PyEnvironment)
        return argument

    # noinspection PyTypeChecker
    def add(self, int_vars: List[PyVar] = None, real_vars: List[PyVar] = None):
        if int_vars:
            i_size = len(int_vars)
            typ: Type = c_char_p * i_size
            int_arr = typ(*(x._as_parameter_ for x in int_vars))
        else:
            i_size = 0
            int_arr = None
        if real_vars:
            r_size = len(real_vars)
            typ: Type = c_char_p * r_size
            real_arr = typ(*(x._as_parameter_ for x in real_vars))
        else:
            r_size = 0
            real_arr = None
        self.environment = libapron.ap_environment_add(self, int_arr, i_size, real_arr, r_size)
        if not self.environment:
            raise ValueError('clashing variable names')
        return self

    # noinspection PyTypeChecker
    def rename(self, old_vars: List[PyVar], new_vars: List[PyVar]):
        o_size = len(old_vars)
        n_size = len(new_vars)
        assert o_size == n_size
        old_typ: Type = c_char_p * o_size
        old_arr = old_typ(*(x._as_parameter_ for x in old_vars))
        new_typ: Type = c_char_p * o_size
        new_arr = new_typ(*(x._as_parameter_ for x in new_vars))
        p = DimPerm()
        self.environment = libapron.ap_environment_rename(self, old_arr, new_arr, o_size, p)
        if not self.environment:
            raise ValueError('invalid renaming')
        return self

    # noinspection PyTypeChecker
    def remove(self, del_vars: List[PyVar] = None):
        if del_vars:
            size = len(del_vars)
            typ: Type = c_char_p * size
            arr = typ(*(x._as_parameter_ for x in del_vars))
        else:
            size = 0
            arr = None
        self.environment = libapron.ap_environment_remove(self, arr, size)
        if not self.environment:
            raise ValueError('non-existing variable(s)')
        return self

    def __repr__(self):
        return str(self.environment.contents)

    def __len__(self):
        return self.environment.contents.intdim + self.environment.contents.realdim

    def __contains__(self, item: 'PyVar'):
        assert isinstance(item, PyVar)
        return libapron.ap_environment_dim_of_var(self, item) != AP_DIM_MAX

    def __lt__(self, other: 'PyEnvironment'):
        assert isinstance(other, PyEnvironment)
        return libapron.ap_environment_compare(self, other) == -1

    def __le__(self, other: 'PyEnvironment'):
        assert isinstance(other, PyEnvironment)
        return self.__lt__(other) or self.__eq__(other)

    def __eq__(self, other: 'PyEnvironment'):
        assert isinstance(other, PyEnvironment)
        return libapron.ap_environment_compare(self, other) == 0

    def __ne__(self, other: 'PyEnvironment'):
        assert isinstance(other, PyEnvironment)
        return not self.__eq__(other)

    def __ge__(self, other: 'PyEnvironment'):
        assert isinstance(other, PyEnvironment)
        return self.__gt__(other) or self.__eq__(other)

    def __gt__(self, other: 'PyEnvironment'):
        assert isinstance(other, PyEnvironment)
        return libapron.ap_environment_compare(self, other) == 1

    def __or__(self, other: 'PyEnvironment') -> 'PyEnvironment':
        assert isinstance(other, PyEnvironment)
        d1 = DimChange()
        d2 = DimChange()
        environment = libapron.ap_environment_lce(self, other, byref(d1), byref(d2))
        if not environment:
            raise ValueError('incompatible environments')
        return PyEnvironment(environment)

    def union(self, other: 'PyEnvironment') -> 'PyEnvironment':
        assert isinstance(other, PyEnvironment)
        return self.__or__(other)


pyvar_p = POINTER(c_char_p)
libapron.ap_environment_alloc_empty.restype = POINTER(Environment)
libapron.ap_environment_alloc.argtypes = [pyvar_p, c_size_t, pyvar_p, c_size_t]
libapron.ap_environment_alloc.restype = POINTER(Environment)
libapron.ap_environment_add.argtypes = [PyEnvironment, pyvar_p, c_size_t, pyvar_p, c_size_t]
libapron.ap_environment_add.restype = POINTER(Environment)
libapron.ap_environment_remove.argtypes = [PyEnvironment, pyvar_p, c_size_t]
libapron.ap_environment_remove.restype = POINTER(Environment)
libapron.ap_environment_compare.argtypes = [PyEnvironment, PyEnvironment]
libapron.ap_environment_dim_of_var.argtypes = [PyEnvironment, PyVar]
libapron.ap_environment_dim_of_var.restype = Dim
dimchange_p = POINTER(DimChange)
libapron.ap_environment_lce.argtypes = [PyEnvironment, PyEnvironment, dimchange_p, dimchange_p]
libapron.ap_environment_lce.restype = POINTER(Environment)
dimperm_p = POINTER(DimPerm)
libapron.ap_environment_rename.argtypes = [PyEnvironment, pyvar_p, pyvar_p, c_size_t, dimperm_p]
libapron.ap_environment_rename.restype = POINTER(Environment)



"""
C DLLs
======

:Author: Caterina Urban
"""
from ctypes import util, CDLL, c_void_p

libc = CDLL(util.find_library('c'))
cstdout = c_void_p.in_dll(libc, '__stdoutp')

libt1pD = CDLL(util.find_library('libt1pD.so'))
libt1pMPQ = CDLL(util.find_library('libt1pMPQ.so'))
libt1pMPFR = CDLL(util.find_library('libt1pMPFR.so'))

libpolkaMPQ = CDLL(util.find_library('libpolkaMPQ.so'))
libpolkaRll = CDLL(util.find_library('libpolkaRll.so'))

liboctD = CDLL(util.find_library('liboctD.so'))
liboctMPQ = CDLL(util.find_library('liboctMPQ.so'))

libboxD = CDLL(util.find_library('libboxD.so'))
libboxMPQ = CDLL(util.find_library('libboxMPQ.so'))
libboxMPFR = CDLL(util.find_library('libboxMPFR.so'))

libapron = CDLL(util.find_library('libapron.so'))

libgmp = CDLL(util.find_library('gmp'))
libmpfr = CDLL(util.find_library('mpfr'))



"""
APRON Abstract Values (Level 0)
===============================

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER
from ctypes import c_void_p

from apronpy.manager import Manager


class Abstract0(Structure):
    """
    struct ap_abstract0_t {
      void* value;       /* Abstract value of the underlying library */
      ap_manager_t* man; /* Used to identify the effective type of value */
    };
    """

    _fields_ = [
        ('value', c_void_p),
        ('man', POINTER(Manager))
    ]



"""
APRON Manager

:Author: Caterina Urban
"""
from _ctypes import Structure, POINTER
from ctypes import c_char_p, c_void_p, CFUNCTYPE, c_size_t, c_int, c_bool, c_uint
from enum import IntEnum


class FunId(IntEnum):
    """
    typedef enum ap_funid_t {
      AP_FUNID_UNKNOWN,
      AP_FUNID_COPY,
      AP_FUNID_FREE,
      AP_FUNID_ASIZE, /* For avoiding name conflict with AP_FUNID_SIZE */
      AP_FUNID_MINIMIZE,
      AP_FUNID_CANONICALIZE,
      AP_FUNID_HASH,
      AP_FUNID_APPROXIMATE,
      AP_FUNID_FPRINT,
      AP_FUNID_FPRINTDIFF,
      AP_FUNID_FDUMP,
      AP_FUNID_SERIALIZE_RAW,
      AP_FUNID_DESERIALIZE_RAW,
      AP_FUNID_BOTTOM,
      AP_FUNID_TOP,
      AP_FUNID_OF_BOX,
      AP_FUNID_DIMENSION,
      AP_FUNID_IS_BOTTOM,
      AP_FUNID_IS_TOP,
      AP_FUNID_IS_LEQ,
      AP_FUNID_IS_EQ,
      AP_FUNID_IS_DIMENSION_UNCONSTRAINED,
      AP_FUNID_SAT_INTERVAL,
      AP_FUNID_SAT_LINCONS,
      AP_FUNID_SAT_TCONS,
      AP_FUNID_BOUND_DIMENSION,
      AP_FUNID_BOUND_LINEXPR,
      AP_FUNID_BOUND_TEXPR,
      AP_FUNID_TO_BOX,
      AP_FUNID_TO_LINCONS_ARRAY,
      AP_FUNID_TO_TCONS_ARRAY,
      AP_FUNID_TO_GENERATOR_ARRAY,
      AP_FUNID_MEET,
      AP_FUNID_MEET_ARRAY,
      AP_FUNID_MEET_LINCONS_ARRAY,
      AP_FUNID_MEET_TCONS_ARRAY,
      AP_FUNID_JOIN,
      AP_FUNID_JOIN_ARRAY,
      AP_FUNID_ADD_RAY_ARRAY,
      AP_FUNID_ASSIGN_LINEXPR_ARRAY,
      AP_FUNID_SUBSTITUTE_LINEXPR_ARRAY,
      AP_FUNID_ASSIGN_TEXPR_ARRAY,
      AP_FUNID_SUBSTITUTE_TEXPR_ARRAY,
      AP_FUNID_ADD_DIMENSIONS,
      AP_FUNID_REMOVE_DIMENSIONS,
      AP_FUNID_PERMUTE_DIMENSIONS,
      AP_FUNID_FORGET_ARRAY,
      AP_FUNID_EXPAND,
      AP_FUNID_FOLD,
      AP_FUNID_WIDENING,
      AP_FUNID_CLOSURE,
      AP_FUNID_SIZE,
      AP_FUNID_CHANGE_ENVIRONMENT,
      AP_FUNID_RENAME_ARRAY,
      AP_FUNID_SIZE2
    } ap_funid_t;
    """
    AP_FUNID_UNKNOWN = 0
    AP_FUNID_COPY = 1
    AP_FUNID_FREE = 2
    AP_FUNID_ASIZE = 3
    AP_FUNID_MINIMIZE = 4
    AP_FUNID_CANONICALIZE = 5
    AP_FUNID_HASH = 6
    AP_FUNID_APPROXIMATE = 7
    AP_FUNID_FPRINT = 8
    AP_FUNID_FPRINTDIFF = 9
    AP_FUNID_FDUMP = 10
    AP_FUNID_SERIALIZE_RAW = 11
    AP_FUNID_DESERIALIZE_RAW = 12
    AP_FUNID_BOTTOM = 13
    AP_FUNID_TOP = 14
    AP_FUNID_OF_BOX = 15
    AP_FUNID_DIMENSION = 16
    AP_FUNID_IS_BOTTOM = 17
    AP_FUNID_IS_TOP = 18
    AP_FUNID_IS_LEQ = 19
    AP_FUNID_IS_EQ = 20
    AP_FUNID_IS_DIMENSION_UNCONSTRAINED = 21
    AP_FUNID_SAT_INTERVAL = 22
    AP_FUNID_SAT_LINCONS = 23
    AP_FUNID_SAT_TCONS = 24
    AP_FUNID_BOUND_DIMENSION = 25
    AP_FUNID_BOUND_LINEXPR = 26
    AP_FUNID_BOUND_TEXPR = 27
    AP_FUNID_TO_BOX = 28
    AP_FUNID_TO_LINCONS_ARRAY = 29
    AP_FUNID_TO_TCONS_ARRAY = 30
    AP_FUNID_TO_GENERATOR_ARRAY = 31
    AP_FUNID_MEET = 32
    AP_FUNID_MEET_ARRAY = 33
    AP_FUNID_MEET_LINCONS_ARRAY = 34
    AP_FUNID_MEET_TCONS_ARRAY = 35
    AP_FUNID_JOIN = 36
    AP_FUNID_JOIN_ARRAY = 37
    AP_FUNID_ADD_RAY_ARRAY = 38
    AP_FUNID_ASSIGN_LINEXPR_ARRAY = 39
    AP_FUNID_SUBSTITUTE_LINEXPR_ARRAY = 40
    AP_FUNID_ASSIGN_TEXPR_ARRAY = 41
    AP_FUNID_SUBSTITUTE_TEXPR_ARRAY = 42
    AP_FUNID_ADD_DIMENSIONS = 43
    AP_FUNID_REMOVE_DIMENSIONS = 44
    AP_FUNID_PERMUTE_DIMENSIONS = 45
    AP_FUNID_FORGET_ARRAY = 46
    AP_FUNID_EXPAND = 47
    AP_FUNID_FOLD = 48
    AP_FUNID_WIDENING = 49
    AP_FUNID_CLOSURE = 50
    AP_FUNID_SIZE = 51
    AP_FUNID_CHANGE_ENVIRONMENT = 52
    AP_FUNID_RENAME_ARRAY = 53
    AP_FUNID_SIZE2 = 54


class Exc(IntEnum):
    """
    typedef enum ap_exc_t {
      AP_EXC_NONE,             /* no exception detected */
      AP_EXC_TIMEOUT,          /* timeout detected */
      AP_EXC_OUT_OF_SPACE,     /* out of space detected */
      AP_EXC_OVERFLOW,         /* magnitude overflow detected */
      AP_EXC_INVALID_ARGUMENT, /* invalid arguments */
      AP_EXC_NOT_IMPLEMENTED,  /* not implemented */
      AP_EXC_SIZE
    } ap_exc_t;
    """
    AP_EXC_NONE = 0
    AP_EXC_TIMEOUT = 1
    AP_EXC_OUT_OF_SPACE = 2
    AP_EXC_OVERFLOW = 3
    AP_EXC_INVALID_ARGUMENT = 4
    AP_EXC_NOT_IMPLEMENTED = 5
    AP_EXC_SIZE = 6


class ExcLog(Structure):
    """
    typedef struct ap_exclog_t {
      ap_exc_t exn;
      ap_funid_t funid;
      char* msg;                   /* dynamically allocated */
      struct ap_exclog_t* tail;
    } ap_exclog_t;
    """
    pass


ExcLog._fields_ = [
        ('exn', c_uint),
        ('funid', c_uint),
        ('msg', c_char_p),
        ('tail', POINTER(ExcLog))
    ]


class Result(Structure):
    """
    typedef struct ap_result_t {
      ap_exclog_t* exclog; /* history of exceptions */
      ap_exc_t exn;        /* exception for the last called function */
      bool flag_exact;  /* result is mathematically exact or don't know */
      bool flag_best;   /* result is best correct approximation or don't know */
    } ap_result_t;
    """

    _fields_ = [
        ('exclog', POINTER(ExcLog)),
        ('exn', c_uint),
        ('flag_exact', c_bool),
        ('flag_best', c_bool)
    ]


class FunOpt(Structure):
    """
    typedef struct ap_funopt_t {
      int algorithm;
      /* Algorithm selection:
         - 0 is default algorithm;
         - MAX_INT is most accurate available;
         - MIN_INT is most efficient available;
         - otherwise, no accuracy or speed meaning
      */
      size_t timeout; /* unit !? */
      /* Above the given computation time, the function may abort with the
         exception flag flag_time_out on.
      */
      size_t max_object_size; /* in abstract object size unit. */
      /* If during the computation, the size of some object reach this limit, the
         function may abort with the exception flag flag_out_of_space on.
      */
      bool flag_exact_wanted;
      /* return information about exactitude if possible
      */
      bool flag_best_wanted;
      /* return information about best correct approximation if possible
      */
    } ap_funopt_t;
    """

    _fields_ = [
        ('algorithm', c_int),
        ('timeout', c_size_t),
        ('max_object_size', c_size_t),
        ('flag_exact_wanted', c_bool),
        ('flag_best_wanted', c_bool)
    ]


# noinspection PyTypeChecker
class Option(Structure):
    """
    typedef struct ap_option_t {
      ap_funopt_t funopt[AP_FUNID_SIZE];
      bool abort_if_exception[AP_EXC_SIZE];
      ap_scalar_discr_t scalar_discr; /* Preferred type for scalars */
    } ap_option_t;
    """

    _fields_ = [
        ('funopt', FunOpt * FunId.AP_FUNID_SIZE),
        ('abort_if_exception', c_bool * Exc.AP_EXC_SIZE),
        ('scalar_discr', c_uint)
    ]


# noinspection PyTypeChecker
class Manager(Structure):
    """
    typedef struct ap_manager_t {
      const char* library;           /* name of the effective library */
      const char* version;           /* version of the effective library */
      void* internal;                /* library dependent,
                                        should be different for each thread
                                        (working space) */
      void* funptr[AP_FUNID_SIZE];   /* Array of function pointers,
                                        initialized by the effective library */
      ap_option_t option;            /* Options (in) */
      ap_result_t result;            /* Exceptions and other indications (out) */
      void (*internal_free)(void*);  /* deallocation function for internal */
      size_t count;                  /* reference counter */
    } ap_manager_t;
    """

    _fields_ = [
        ('library', c_char_p),
        ('version', c_char_p),
        ('internal', c_void_p),
        ('funpts', c_void_p * FunId.AP_FUNID_SIZE),
        ('option', Option),
        ('result', Result),
        ('internal_free', CFUNCTYPE(None, c_void_p)),
        ('count', c_size_t)
    ]



from setuptools import setup, find_packages

setup(
    name='apronpy',
    version='0.9',
    author='Caterina Urban',
    author_email='caterina.urban@gmail.com',
    description='Python Interface for the APRON Numerical Abstract Domain Library',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/caterinaurban/apronpy',
    packages=find_packages(),
)
